/// <reference path="./eServiceFormHelpers.js"/>

/**
 * Cached result from the BookingAPI for BookingApi_MakeApiCall
 */
var selectTime_OfficeAvailabilities = [];
var selectTime_Office = null;
var calendarApi_Date = null;
var calendarApi_Time = null;

var loadingOfficeFetch = false;
var esdc_officesFetch = [];

/**
 * Azure Maps Services
 */
var mapService = null;
var searchService = null;

/**
 * Configures the Booking Form with all the required validation and Views
 * @param {{}} options Optional Parameters to control the Loading of the Form
 * @returns {Object.<string, InitializeLogic>} The Booking Initialize Logic
 */
function CreateBookingInitializationLogic(options) {
   if (options == null || options == undefined) { options = {}; }

   let initalizeLogic = {};

   //#region Begin Initialization
   createInitializeLogic(initalizeLogic, "esdc_languageofservice", function () {
      // Create the Info section that pops up
      let infoAlert = createAlert("esdc_otherspokenlanguage_infosection", "alert-info", getTranslatedText(
         `If you need service in a language other than English or French, please try to bring a family member or a friend with you that can help interpret for you. Not all Service Canada Centres offer an interpretation service.`,
         `Si vous avez besoin d'un service dans une langue autre que l'anglais ou le français, essayer d'amener avec vous un membre de votre famille ou un ami qui pourrait traduire pour vous. Les Centres Service Canada n'offrent pas tous un service d'interprétation.`
      ), false, {
         backingId: "esdc_otherspokenlanguage",
         describes: {
            fields: ["esdc_otherspokenlanguage"]
         }
      });
      $(infoAlert).insertBefore($("#esdc_otherspokenlanguage"));

      $("#esdc_languageofservice").change(function () {
         let esdc_languageofservice = $(this).val();
         if (esdc_languageofservice == "564190002") {
            $("#esdc_otherspokenlanguage").parents("td").show();
            $("#esdc_otherspokenlanguage_infosection").show();
            addRequiredValidator("esdc_otherspokenlanguage", null, null);
            configureRequiredLabels({
               fields: ["esdc_otherspokenlanguage"]
            });
         }
         else {
            $("#esdc_otherspokenlanguage_infosection").hide();
            $("#esdc_otherspokenlanguage").parents("td").hide();
            removeRequiredValidator("esdc_otherspokenlanguage");
         }
      });

      $("#esdc_languageofservice").change();
   });

   createInitializeLogic(initalizeLogic, "esdc_firstname", function () {
      $("#esdc_firstname").attr("autocomplete", "given-name");
   });
   createInitializeLogic(initalizeLogic, "esdc_lastname", function () {
      $("#esdc_lastname").attr("autocomplete", "family-name");
   });

   createInitializeLogic(initalizeLogic, "esdc_provinceterritorystateid", async function () {
      let selectedProvinceTerritoryState = getLookupField("esdc_provinceterritorystateid");
      $("#esdc_provinceterritorystateid").children().remove();

      let addressMetadata = await loadAddressMetadata(null);

      let provinces = addressMetadata.provinceTerritoryStates.value.filter(x => {
         /* Search for Canadian Provinces Canada */
         if (x._esdc_countryid_value == "0cf5389e-97ae-eb11-8236-000d3af4bfc3") {
            return true;
         }
         return false;
      }).map(x => {
         return {
            text: getTranslatedText(x.esdc_nameenglish, x.esdc_namefrench),
            value: x.esdc_provinceterritorystateid,
            groupId: x.esdc_groupkey,
            attributes: [
               { key: "data-name-english", value: x.esdc_nameenglish },
               { key: "data-name-french", value: x.esdc_namefrench },
               { key: "data-countryid", value: x._esdc_countryid_value }
            ]
         }
      });

      let provincesGroup = createOptionList(provinces, {
         text: getTranslatedText("Select your Province/Territory", "Sélectionnez votre province/territoire"),
         value: ""
      });

      $("#esdc_provinceterritorystateid").append(provincesGroup);

      if (selectedProvinceTerritoryState != null) {
         setLookupField("esdc_provinceterritorystateid", selectedProvinceTerritoryState);
      }
   });

   createInitializeLogic(initalizeLogic, "esdc_primarytelephonenumber", function () {
      $("#esdc_primarytelephonenumber").attr("type", "tel");
      $("#esdc_primarytelephonenumber").attr("autocomplete", "tel");
   });
   createInitializeLogic(initalizeLogic, "esdc_mobiletelephonenumber", function () {
      $("#esdc_mobiletelephonenumber").attr("type", "tel");

      // Configure the Copy Button for Mobile Phone Number
      let copyEnabled = $("#esdc_mobiletelephonenumber").attr("data-copyButtonEnabled");
      if (copyEnabled == "true") {
         let daytimeLabelText = getLabelText("esdc_primarytelephonenumber");
         let copyLabelText = daytimeLabelText.substring(daytimeLabelText.indexOf("(") + 1, daytimeLabelText.indexOf(")"));
         createInputGroup("esdc_mobiletelephonenumber", {
            buttons: [
               {
                  content: {
                     text: getTranslatedText("Copy {0}", "Copier {0}").replace("{0}", copyLabelText)
                  },
                  buttonType: "btn-default",
                  position: "right",
                  attributes: [
                     { "key": "id", value: "CopyTelephoneDaytimeButton" },
                     { "key": "title", value: getTranslatedText("Copy {0}", "Copier {0}").replace("{0}", daytimeLabelText) },
                     { "key": "data-sourceId", value: "esdc_primarytelephonenumber" },
                     { "key": "data-backingId", value: "esdc_mobiletelephonenumber" },
                  ],
                  action: function () {
                     let sourceId = $(this).attr("data-sourceId");
                     let backingId = $(this).attr("data-backingId");

                     let daytimeValue = $("#" + sourceId).val();
                     $("#" + backingId).val(daytimeValue);
                  }
               }
            ]
         });
      }
   });
   createInitializeLogic(initalizeLogic, "esdc_emailaddress", function () {
      $("#esdc_emailaddress").attr("autocomplete", "email");
      $("#esdc_emailaddress").parents("td").find(".description").attr("data-configureTooltips-hidden", false);

      // Configure the Required Labels
      configureRecommendedLabels({ fields: ["esdc_emailaddress"] });
      configureRequiredLabels({ fields: ["esdc_emailaddress"] });
      $("#esdc_emailaddress_label").find(".requiredText").hide(); //hide required label
      $("#esdc_emailaddress_label").find(".recommendedText").hide();//hide recommended label

      let esdc_bookingtype_value = $("#esdc_bookingtype").val();
      if ((esdc_bookingtype_value == "564190001") /* Passport */ || (isNullOrWhitespace(esdc_bookingtype_value))) {
         $("#esdc_emailaddress_label").find(".recommendedText").show();//show recommended label
         removeRequiredValidator("esdc_emailaddress");
      }
      else if (esdc_bookingtype_value == "564190000" /* Biometrics */) {
         addRequiredValidator("esdc_emailaddress", null, { allowDuplicates: false });
         $("#esdc_emailaddress_label").find(".requiredText").show();//show required label
      }
   });
   createInitializeLogic(initalizeLogic, "esdc_postalcodezip", function () {
      this.jquery.change(function () {
         let selectedValue = $(this).val()?.toUpperCase();
         $(this).val(selectedValue);
      });
   });

   createInitializeLogic(initalizeLogic, "esdc_earliesttraveldate", function () {
      // Reparent the td
      let newDiv = document.createElement("div");
      let existingControl = $("#esdc_earliesttraveldate").parents("td");
      $(newDiv).append(existingControl.children());
      $(existingControl).append(newDiv);

      // Split the Date Field
      splitDateInputControl("esdc_earliesttraveldate", true, true, {
         compileMode: "date-format",
         allowShortYear: false,
         allowFutureDates: true,
         allowFutureYears: true,
         allowPastDates: false,
         nullBackingOnInvalid: true,
         year: {
            incrementingList: true,
            max: new Date().getFullYear() + 10
         }
      });

      // Configure the Minimum Validator
      createValidator("esdc_earliesttraveldate", null, "MinimumDateSelection", "esdc_earliesttraveldate", false, getTranslatedText(
         "If you require a passport in the next 2 days, call the Passport Call Centre at 1-800-567-6868.",
         "Si vous avez besoin d'un passeport dans les prochains 2 jours, appelez le Centre d'appels pour passeports au 1-800-567-6868."
      ), function () {
         let esdc_bookingtype_value = $("#esdc_bookingtype").val();
         if (esdc_bookingtype_value != "564190001" /* Passport */) { return true; }

         let selectedValue = $("#esdc_earliesttraveldate").val();
         if (!isNullOrWhitespace(selectedValue)) {
            let differenceDays = moment.utc(selectedValue).diff(moment.utc().startOf("day"), "day");
            if (differenceDays <= 2) return false;
         }

         return true;
      });

      // Add the unncessesary N/A option
      let esdc_earliesttraveldateNA_value = sessionLoadObject("esdc_earliesttraveldate_na_value");
      let esdc_earliesttraveldateNA = createPortalField("esdc_earliesttraveldate_na", {
         labelText: getTranslatedText(
            "Do you want to provide your earliest travel date?",
            "Souhaitez-vous indiquer votre première date de voyage ?"
         )
      }, true, {
         elementType: "select",
         defaultValue: (esdc_earliesttraveldateNA_value == null) ? "1" : esdc_earliesttraveldateNA_value,
         options: [
            { value: 0, text: getTranslatedText("No", "Non") },
            { value: 1, text: getTranslatedText("Yes", "Oui") }
         ],
         attributes: [
            { key: "data-backingid", value: "esdc_earliesttraveldate" }
         ],
         handlers: [
            {
               eventName: "change",
               handler: function () {
                  let thisId = $(this).attr("id");
                  let selectedValue = $(this).val();
                  sessionSaveObject(thisId + "_value", selectedValue);

                  let backingId = $(this).attr("data-backingid");
                  if (selectedValue == "1") {
                     $("#" + backingId).parent().parent().show();
                     addRequiredValidator(backingId, null);
                  }
                  else {
                     $("#" + backingId).parent().parent().hide();
                     removeRequiredValidator(backingId);
                  }
               }
            }
         ]
      });
      $(esdc_earliesttraveldateNA).insertBefore(newDiv);
   });

   createInitializeLogic(initalizeLogic, "esdc_earliesttraveldaterange", function () {
      if (this.jquery.attr("data-disablechange") == true) { console.log("Initializer Disabled (esdc_earliesttraveldaterange)"); return; }

      // Create the Warning for if we are between 2-9 days
      let warningAlert = createAlert("esdc_earliesttraveldaterange_2to9days_warning", "alert-warning", getTranslatedText(
         `Proof of travel or need may be required. <a href="https://www.canada.ca/en/immigration-refugees-citizenship/services/canadian-passports/fees/passport-canada.html" target="_blank">Additional fees</a> apply.`,
         `Une preuve de voyage ou de besoin pourrait être requise. Des <a href="https://www.canada.ca/fr/immigration-refugies-citoyennete/services/passeports-canadiens/frais/passeport-canada.html" target="_blank">frais supplémentaires</a> s'appliquent.`
      ), false, {
         backingId: "esdc_earliesttraveldaterange",
         describes: {
            fields: ["esdc_earliesttraveldaterange"]
         }
      });
      $(warningAlert).insertBefore($("#esdc_earliesttraveldaterange"));

      // Create the Error if we are under 2 days or less
      let under2DaysErrorText = getTranslatedText(
         `If you need a passport in the next 2 business days, visit a <a href='https://ircc.canada.ca/english/passport/map/map.asp?po=1#table1caption' target='_blank'>specialized passport site (Service Canada Centre - Passport Service)</a> that offers urgent pick-up service.<br/>Proof of travel or need is required. <a href="https://www.canada.ca/en/immigration-refugees-citizenship/services/canadian-passports/fees/passport-canada.html" target="_blank">Additional fees</a> apply.`,
         `Si vous avez besoin d'un passeport dans les 2 prochains jours ouvrables, visitez un <a href='https://ircc.canada.ca/francais/passeport/map/carte.asp?po=1#table1caption' target='_blank'>site spécialisé dans les passeports (Centre Service Canada - Service de passeport)</a> qui offre le service de retrait urgent.<br/>Une preuve de voyage ou de besoin est requise. Des <a href="https://www.canada.ca/fr/immigration-refugies-citoyennete/services/passeports-canadiens/frais/passeport-canada.html" target="_blank">frais supplémentaires</a> s'appliquent.`
      );
      let errorAlert = createAlert("esdc_earliesttraveldaterange_0to2days_error", "alert-danger", under2DaysErrorText, false, {
         backingId: "esdc_earliesttraveldaterange",
         describes: {
            fields: ["esdc_earliesttraveldaterange"]
         }
      });
      $(errorAlert).insertBefore($("#esdc_earliesttraveldaterange"));

      createValidator("esdc_earliesttraveldaterange", null, "MinimumDateSelection", "esdc_earliesttraveldaterange", false, under2DaysErrorText, function () {
         let esdc_bookingtype_value = $("#esdc_bookingtype").val();
         if (esdc_bookingtype_value != "564190001" /* Passport */) { return true; }

         let selectedValue = $("#esdc_earliesttraveldaterange").val();
         if (selectedValue == "564190001") /* 2 business days or less */ {
            return false;
         }

         return true;
      });

      // Apply the change handler for the Warning/Error
      $("#esdc_earliesttraveldaterange").change(function () {
         let selectedValue = $("#esdc_earliesttraveldaterange").val();
         let esdc_bookingtype_value = $("#esdc_bookingtype").val();

         // Hide all of the alerts
         $("#esdc_earliesttraveldaterange_2to9days_warning").hide();
         $("#esdc_earliesttraveldaterange_0to2days_error").hide();
         $("#esdc_earliesttraveldaterange_6to25days_warning").hide();

         // Determine whether the RA Checklist is visible, and which Errors/Warnings to use
         let activeAlertId = "";
         let raVisible = true;
         if (selectedValue == "564190000" /* No Travel Date */) { raVisible = true; }
         else if (selectedValue == "564190001") /* 2 Days or less */ { raVisible = false; activeAlertId = "esdc_earliesttraveldaterange_0to2days_error"; }
         else if (selectedValue == "564190002") /* 3-9 Business Days */ { raVisible = false; activeAlertId = "esdc_earliesttraveldaterange_2to9days_warning"; }
         else if (selectedValue == "564190003") /* 3-9 Business Days */ { raVisible = true; activeAlertId = "esdc_earliesttraveldaterange_6to25days_warning"; }
         else if (selectedValue == "564190004") /* 25+ Business Days */ { raVisible = true; }
         else /* (isNullOrWhitespace(selectedValue)) */ { raVisible = true; }

         // Apply the Checklist Visibility
         if (raVisible) {
            //Show the RA checklsit field and tooltip
            $("#esdc_passportraapplicable").parents("td").show();

            //Reinstate the field validation for the RA checlist field
            if (esdc_bookingtype_value == "564190001") /* Passport */ {
               addRequiredValidator("esdc_passportraapplicable", null, { allowDuplicates: false });
            }
         }
         else {
            //Set RA checklist value to NULL
            $("#esdc_passportraapplicable").val("");

            //Allow the user to move forward by bypassing the RA checklist field validation
            removeRequiredValidator("esdc_passportraapplicable");

            //Hide the RA checklist field and tooltip
            $("#esdc_passportraapplicable").parents("td").hide();
         }

         // Display the correct Alert
         if (!isNullOrWhitespace(activeAlertId)) {
            $("#" + activeAlertId).show();
            setAriaErrorMessage("esdc_earliesttraveldaterange", activeAlertId, true, true);
         }
      });
      $("#esdc_earliesttraveldaterange").change();
   });

   createInitializeLogic(initalizeLogic, "esdc_numberofapplicantsbio", function () {
      convertNumberToRange("esdc_numberofapplicantsbio", 0, 0, 8);
      // Toggle the Tooltip by default
      $("#esdc_numberofapplicantsbio").parents("td").find(".description").attr("data-configureTooltips-hidden", false);
      $("#esdc_numberofapplicantsbio").attr("placeholder", "");

      // Add validator for minimum applicants > 0
      createValidator("esdc_numberofapplicantsbio", null, "MinimumApplications", "esdc_numberofapplicantsbio", false, "",
      function () {
            let esdc_bookingtype_value = $("#esdc_bookingtype").val();
            if (esdc_bookingtype_value == "564190000") /* BIO */ {
               let selectedValue = $("#esdc_numberofapplicantsbio").val();
               if (parseInt(selectedValue) > 0) {
                  return true;
               }
               return false;
            }

            return true;
         }
      );
   });

   createInitializeLogic(initalizeLogic, "esdc_numberofapplicationspassport", function () {
      convertNumberToRange("esdc_numberofapplicationspassport", 0, 0, 6);

      // Add validator for minimum applications > 0
      createValidator("esdc_numberofapplicationspassport", null, "MinimumApplications", "esdc_numberofapplicationspassport", false, "",
         function () {
            let esdc_bookingtype_value = $("#esdc_bookingtype").val();
            if (esdc_bookingtype_value == "564190001") /* Passport */ {
               let selectedValue = $("#esdc_numberofapplicationspassport").val();
               if (parseInt(selectedValue) > 0) {
                  return true;
               }
               return false;
            }

            return true;
         }
      );
   });

   createInitializeLogic(initalizeLogic, "esdc_bilapplicationnumber", function () {
      // Toggle the Tooltip by default
      $("#esdc_bilapplicationnumber").parents("td").find(".description").attr("data-configureTooltips-hidden", false);
      $("#esdc_bilapplicationnumber").attr("placeholder", ""); // getTranslatedText('AB1234567 or A12345678', 'AB1234567 ou A12345678')
   });

   createInitializeLogic(initalizeLogic, "esdc_passportraapplicable", function () {
      if (this.jquery.attr("data-disablechange") == true) { console.log("Initializer Disabled (esdc_passportraapplicable)"); return; }

      let esdc_passportraapplicableTd = $("#esdc_passportraapplicable").parents("td");
      esdc_passportraapplicableTd.find(".description").attr("data-configureTooltips-hidden", false);
      let RAchecklistdefault = $("#esdc_passportraapplicable option:eq(0)");
      RAchecklistdefault.removeAttr("aria-label");
      RAchecklistdefault.removeAttr("label");
      RAchecklistdefault.text(getTranslatedText("Select Yes/No", "Sélectionnez Oui/Non"));

      // Because of how this question is worded...
      // ie: The Field is named Positively, and the question prompt is negative
      // ... We are going to switch the Text on the Yes and No Values such that the data in CRM
      // Matches up as expected. We switch the text so that when the User hits the back or previous buttons
      // The data continues to match up
      let options = $("#esdc_passportraapplicable").children();
      let noOption = options.eq(1);
      let yesOption = options.eq(2);

      let tmpNoText = noOption.text();
      noOption.text(yesOption.text());
      yesOption.text(tmpNoText);
   });

   createInitializeLogic(initalizeLogic, "esdc_bookingtype", function () {
      if ($("#esdc_bookingtype").attr("data-disablechange") != true) {
         $("#esdc_bookingtype").change(function () {
            // Reset back to default
            let bioSection = $("[data-name='step0_tab_section_bio_1']").parents("fieldset");
            bioSection.hide();

            let passportSection = $("[data-name='step0_tab_section_passport_1']").parents("fieldset");
            passportSection.hide();

            // Remove Validators
            removeRequiredValidator("esdc_bilapplicationnumber");
            removeRequiredValidator("esdc_numberofapplicationspassport");
            removeRequiredValidator("esdc_numberofapplicantsbio");
            removeRequiredValidator("esdc_privateroomrequired");
            removeRequiredValidator("esdc_passportraapplicable");
            removeRequiredValidator("esdc_earliesttraveldate");
            removeRequiredValidator("esdc_earliesttraveldaterange");

            //Grab the current booking type value picked
            let esdc_bookingtype_value = $(this).val();

            // Configure individual sections
            if (esdc_bookingtype_value == "564190000") /* BIO */ {
               passportSection.find("input,select,textarea").val(null);
               $("#esdc_numberofapplicationspassport").val(0).change();

               $("[data-name='step0_tab_section_bio_1']").parents("fieldset").show();
               addRequiredValidator("esdc_bilapplicationnumber");
               addRequiredValidator("esdc_numberofapplicantsbio");
               addRequiredValidator("esdc_privateroomrequired");
               configureRequiredLabels({
                  fields: ["esdc_bilapplicationnumber", "esdc_numberofapplicantsbio"]
               });

            }
            else if (esdc_bookingtype_value == "564190001") /* Passport */ {
               bioSection.find("input,select,textarea").val(null);
               $("#esdc_numberofapplicantsbio").val(0).change();

               $("[data-name='step0_tab_section_passport_1']").parents("fieldset").show();
               addRequiredValidator("esdc_numberofapplicationspassport");
               addRequiredValidator("esdc_earliesttraveldaterange");
               addRequiredValidator("esdc_passportraapplicable", null, { allowDuplicates: false });
               configureRequiredLabels({
                  fields: ["esdc_numberofapplicationspassport", "esdc_earliesttraveldaterange", "esdc_passportraapplicable"]
               });

               $("#esdc_earliesttraveldate_na").change();
            }
         });

         $("#esdc_bookingtype").change();
      }
   });

   createInitializeLogic(initalizeLogic, "esdc_privacystatementaccepted", function () {
      // Search the Session and the Querystring for the Privacy Statement Acceptance
      let privacyStatementSession = sessionLoadObject("booking-privacy");
      if (privacyStatementSession == "true") {
         $("#esdc_privacystatementaccepted").val(1);
      }

      const queryString = location.search;
      console.log("Query String:" + queryString);
      if (!isNullOrWhitespace(queryString)) {
         const query = parseQuery(location.search);
         if (query["booking-privacy"]) {
            $("#esdc_privacystatementaccepted").val(1);
         }
      }

      // Get the Value of the Privacy Statement
      let currentValue = $("#esdc_privacystatementaccepted").val();
      if (currentValue == "1") {
         console.log("Privacy Statement Accepted");
      }
      else {
         let locationTmp = window.location.href;
         locationTmp = locationTmp.replace("application/", "");
         console.log("Redirecting to Privacy Page: " + locationTmp);

         try {
            window.location.replace(locationTmp);
         } catch (ex) { }
      }
   });

   createInitializeLogic(initalizeLogic, "esdc_portalbooking", function () {
      // Flag this as a Portal Booking
      try { $("#esdc_portalbooking").val(1); }
      catch (ex) {
         let esdc_portalbooking = document.getElementById("esdc_portalbooking");
         esdc_portalbooking.value = 1;
      }

      createValidator("esdc_portalbooking", null, "TurnOn", "esdc_portalbooking", false, "", function () {
         let esdc_portalbooking = document.getElementById("esdc_portalbooking");
         esdc_portalbooking.value = 1;
         return true;
      })

      return true;
   });

   createInitializeLogic(initalizeLogic, "esdc_browseragent", function () {
      SaveUserAgent("esdc_browseragent");
      // $("#esdc_browseragent").attr("readonly", "readonly");
   });

   createInitializeLogic(initalizeLogic, "esdc_clientianatimezone", function () {
      try {
         var timezone = Intl.DateTimeFormat().resolvedOptions().timeZone;
         $("#esdc_clientianatimezone").val(timezone);
      } catch (ex) { }
   });

   createInitializeLogic(initalizeLogic, "esdc_privateroomrequired", function () {

   });

   createInitializeLogic(initalizeLogic, "esdc_firstdateofavailability", function () {
      $("#esdc_firstdateofavailability_datepicker_description").removeAttr("aria-describedby");
   });

   createInitializeLogic(initalizeLogic, "esdc_accessibilityinformation", function () {
      let esdc_accessibilityinformationTd = $("#esdc_accessibilityinformation").parents("td");
      esdc_accessibilityinformationTd.find(".description").attr("data-configureTooltips-hidden", false);
   });

   createInitializeLogic(initalizeLogic, "step1_tab_section_additionalservices_fieldset", function () {
      createCollapsable(this.jquery, null, {
         defaultState: "open",
         addBreakAfterDetails: true,
         containerType: "div"
      });
   });

   createInitializeLogic(initalizeLogic, "step1_tab_section_findoffice_fieldset", function () {
      // Determine whether to render the subgrid based off if the Additional Offices have been viewed or not
      let sessionViewOnceValue = sessionLoadObject("officeadditionalservice_subgrid_viewonce");
      if (sessionViewOnceValue) { this.jquery.show(); }
      else { this.jquery.hide(); }

      createValidator("closest-offices", $("#step1_tab_section_findoffice_fieldset").attr("aria-label"),
         "Required", "step1_tab_section_findoffice_fieldset", true, getTranslatedText("is a required field.", "est un champ requis."),
         function () {
            let closestOffices = $("#closest-offices");
            if (closestOffices.length <= 0) { return false; }

            let selectedOfficeIds = $("#closest-offices").attr("data-selected-office-id");
            return !isNullOrWhitespace(selectedOfficeIds);
         }
      );

      // Create a container for messaging
      let messagingContainer = $(document.createElement("div"));
      messagingContainer.attr("id", "step1_tab_section_findoffice_fieldset_messaging");
      $(messagingContainer).insertAfter($("#step1_tab_section_findoffice_fieldset").find('legend'));
   });

   createInitializeLogic(initalizeLogic, "step1_tab_section_selecttime_fieldset", function () {
      let lookup = getLookupField("esdc_officeid");

      if (lookup == null) { this.jquery.hide(); }
      else { this.jquery.show(); }

      let tdColumn = this.jquery.find("tr").find("td").first();

      // Create a Progress Indicator to display on the page
      let progressIndicator = createProgressIndicator("selectTime_loading", {
         type: "circle",
         indeterminate: true,
         enabled: true,
         container: {
            position: "relative",
            class: "row",
            text: getTranslatedText(
               "Please wait while availabilities are being calculated for this office...",
               "Veuillez patienter pendant que les disponibilités sont calculées pour ce bureau..."
            )
         },
         style: {
            display: "inline",
            position: "relative",
            left: "0px"
         },
         circle: {
            width: 100,
            height: 100,
            radius: 40
         }
      });
      tdColumn.append(progressIndicator);


      // Create Error message
      let officeNoAvailabilitiesError = createAlert("selectTime_error", "alert-danger", getTranslatedText(
         "This office has no appointments available. Please select a different office.",
         "Ce bureau n'a pas de rendez-vous disponibles. Veuillez sélectionner un autre bureau."
      ), false, {
         role: "alert",
         backingId: "esdc_officeid",
         describes: { fields: ["esdc_officeid"] }
      });
      tdColumn.append(officeNoAvailabilitiesError);

      // Create the Selectors
      // WET Calendar - https://wet-boew.github.io/v4.0-ci/demos/cal-events/cal-events-en.html
      let availabilitySelector = $(document.createElement("div"));
      availabilitySelector.attr("id", "selectTime");
      availabilitySelector.addClass("row");
      availabilitySelector.hide();
      tdColumn.append(availabilitySelector);

      let dateSelectorContainer = $(document.createElement("div"));
      dateSelectorContainer.addClass("col-xs-12");
      dateSelectorContainer.addClass("col-md-6");
      availabilitySelector.append(dateSelectorContainer);

      let timeSelectorContainer = $(document.createElement("div"));
      timeSelectorContainer.addClass("col-xs-12");
      timeSelectorContainer.addClass("col-md-6");
      availabilitySelector.append(timeSelectorContainer);

      // Create Date Selector Components
      let dateSelector = createPortalField("selectTime_date", {
         labelText: getTranslatedText(
            "Date",
            "Date"
         ),

      }, true, {
         containerType: "div",
         elementType: "select",
         defaultValue: null,
         visible: true,
         readonly: false,
         handlers: [
            {
               eventName: "change",
               handler: function () {
                  $("#selectTime_time").children().remove();

                  let selectedValue = $(this).val();
                  if (isNullOrWhitespace(selectedValue)) { console.log("No time selected"); hideAndlockNextButton(); return; }
                  if (selectedValue == -1) { console.log("Default option selected"); hideAndlockNextButton(); return; }

                  $(this).attr("data-selected-datetime-iso", selectedValue);

                  if (!variableExists(selectTime_OfficeAvailabilities)) { console.log("selectTime_OfficeAvailabilities not set"); hideAndlockNextButton(); return; }
                  if (!variableExists(selectTime_Office)) { console.log("selectTime_Office not set"); hideAndlockNextButton(); return; }
                  let groupedTimeslots = groupTimeslotsForDate(selectedValue);

                  let timeslotKeys = Object.keys(groupedTimeslots);
                  let timeOptions = createOptionList(timeslotKeys.map((value, index) => {
                     let localTime12Hour = moment(value, "HH:mm").format("hh:mm A");
                     return {
                        id: value,
                        value: value,
                        text: `${localTime12Hour} (${value})`,
                        attributes: [
                           { key: "data-timeslot-count", value: groupedTimeslots[value].length },
                           { key: "data-time-24hour", value: value },
                           { key: "data-time-12hour", value: localTime12Hour },
                           { key: "data-datetime-iso", value: `${selectedValue}T${value}:00Z` }
                        ]
                     }
                  }), {
                     text: getTranslatedText(
                        "Please select a time slot",
                        "Veuillez sélectionner une plage horaire"
                     )
                  });

                  $("#selectTime_time").append(timeOptions);

                  // Update the color on the Calendar Event
                  if (variableExists(calendarApi_Date)) {
                     // Go to the Date on the Date Calendar
                     try {
                        calendarApi_Date.gotoDate(moment.utc(selectedValue).toDate());
                     } catch (ex) { console.log("There was an error navigating on the Date calendar"); }

                     // Update the color of the selected events
                     let events = calendarApi_Date.getEvents();
                     for (let i = 0; i < events.length; i++) {
                        let event = events[i];
                        if (selectedValue == event.extendedProps.dateKey) {
                           event.setProp("backgroundColor", event.extendedProps.options.selectedColor);
                        }
                        else {
                           event.setProp("backgroundColor", event.extendedProps.options.deselectedColor);
                        }
                     }
                  }

                  if (variableExists(calendarApi_Time)) {
                     // Go to the Date on the Time Calendar
                     try {
                        calendarApi_Time.gotoDate(moment.utc(selectedValue).toDate());
                        calendarApi_Time.refetchEvents();
                     } catch (ex) { console.log("There was an error navigating on the Time calendar"); }
                  }
               }
            }
         ]
      });
      $(dateSelector).addClass("col-xs-12");
      $(dateSelector).attr("id", "selectTime_date_container");
      dateSelectorContainer.append(dateSelector);

      // Create the Input Group for Date
      // createInputGroup("selectTime_date", {
      //    element: dateSelector,
      //    parent: ".control",
      //    buttons: [
      //       {
      //          content: {
      //             glyphicon: {
      //                position: "right",
      //                type: "glyphicon-step-forward",
      //                aria: getTranslatedText("Next Available Time", "Prochaine plage horaire disponible"),
      //                role: "button"
      //             }
      //          },
      //          elementType: "button",
      //          buttonType: "btn-default",
      //          position: "right",
      //          role: "button",
      //          type: "button",
      //          attributes: [
      //             { key: "data-date-id", value: "selectTime_date" },
      //             { key: "data-time-id", value: "selectTime_time" },
      //             { key: "data-date-index", value: 0 },
      //             { key: "data-time-index", value: -1 },
      //             { key: "data-time-length", value: -1 }
      //          ],
      //          action: function () {
      //             console.log("Finding Next Available Timeslot");
      //             let dateIndex = parseInt($(this).attr("data-date-index"));
      //             let timeIndex = parseInt($(this).attr("data-time-index")) + 1;

      //             // Retrieve all the Date Options
      //             let dateOptions = $("#selectTime_date").find("option:not([value=-1])");
      //             let currentDate = dateOptions.eq(dateIndex);
      //             let currentDateValue = currentDate.attr("value");

      //             // Retrieve the Time Slots, wrapping to the next date if overflowed
      //             let timeOptions = groupTimeslotsForDate(currentDateValue);
      //             let timeOptionKeys = Object.keys(timeOptions);
      //             let timeOptionsCount = timeOptionKeys.length;
      //             $(this).attr("data-time-length", timeOptionsCount);

      //             // Wrap the Time and re-retrieve the time options
      //             if (timeIndex >= timeOptionsCount) {
      //                dateIndex++;
      //                if (dateIndex >= dateOptions.length) { dateIndex = 0; }

      //                timeIndex = 0;
      //                currentDate = dateOptions.eq(dateIndex);
      //                currentDateValue = currentDate.attr("value");
      //                timeOptions = groupTimeslotsForDate(currentDateValue);
      //                timeOptionKeys = Object.keys(timeOptions);
      //                timeOptionsCount = timeOptionKeys.length;
      //                $(this).attr("data-time-length", timeOptionsCount);
      //             }

      //             let currentTimeValue = timeOptionKeys[timeIndex];
      //             console.log(`Next Available Time: ${currentDateValue} @ ${currentTimeValue} - dateIndex:${dateIndex}, timeIndex:${timeIndex}`);

      //             $(this).attr("data-date-index", dateIndex);
      //             $(this).attr("data-time-index", timeIndex);
      //             $("#selectTime_date").val(currentDate.attr("value"));
      //             $("#selectTime_date").change();
      //             $("#selectTime_time").val(currentTimeValue);
      //             $("#selectTime_time").change();
      //          }
      //       }
      //    ]
      // });

      let calendarDateContainer = $(document.createElement("div"));
      calendarDateContainer.addClass("col-xs-12");
      calendarDateContainer.attr("id", "selectTime_date_calendar_container");
      calendarDateContainer.attr("aria-hidden", true);
      calendarDateContainer.attr("tabindex", "-1");
      dateSelectorContainer.append(calendarDateContainer);

      let calendarDateDiv = $(document.createElement("div"));
      calendarDateDiv.css("margin-top", "1em");
      calendarDateDiv.css("margin-bottom", "1em");
      calendarDateDiv.attr("aria-hidden", true);
      calendarDateDiv.attr("tabindex", "-1");
      calendarDateContainer.append(calendarDateDiv);

      // Create Time Selector Components
      let timeSelector = createPortalField("selectTime_time", {
         labelText: getTranslatedText(
            "Time",
            "Heure"
         )
      }, true, {
         containerType: "div",
         elementType: "select",
         defaultValue: null,
         visible: true,
         readonly: false,
         handlers: [
            {
               eventName: "change",
               handler: async function () {
                  let selectedDate = $("#selectTime_date").val();
                  let selectedTime = $("#selectTime_time").val();
                  if (isNullOrWhitespace(selectedDate) || isNullOrWhitespace(selectedTime)) {
                     console.log("Date or Time not selected. Form now invalid");
                     hideAndlockNextButton();
                     return;
                  }

                  let selectedTimeOption = $(this).find(`[value='${selectedTime}']`).first();
                  let selectedISO = $(selectedTimeOption).attr("data-datetime-iso");
                  $(this).attr("data-selected-datetime-iso", selectedISO);

                  // Update the color of the selected events
                  if (variableExists(calendarApi_Time)) {
                     let selectedValue = selectedTime;
                     let events = calendarApi_Time.getEvents();
                     for (let i = 0; i < events.length; i++) {
                        let event = events[i];
                        if (selectedValue == event.extendedProps.dateKey) {
                           event.setProp("backgroundColor", event.extendedProps.options.selectedColor);
                        }
                        else {
                           event.setProp("backgroundColor", event.extendedProps.options.deselectedColor);
                        }
                     }
                  }

                  // Enable the Next Button
                  $("#NextButton").removeAttr("disabled");
                  $("#NextButton").show();
               }
            }
         ]
      });
      $(timeSelector).addClass("col-xs-12");
      $(timeSelector).attr("id", "selectTime_time_container");
      timeSelectorContainer.append(timeSelector);

      let calendarTimeContainer = $(document.createElement("div"));
      calendarTimeContainer.addClass("col-xs-12");
      calendarTimeContainer.attr("id", "selectTime_time_calendar_container");
      calendarTimeContainer.attr("aria-hidden", true);
      calendarTimeContainer.attr("tabindex", "-1");
      timeSelectorContainer.append(calendarTimeContainer);

      let calendarTimeDiv = $(document.createElement("div"));
      calendarTimeDiv.css("margin-top", "1em");
      calendarTimeDiv.css("margin-bottom", "1em");
      calendarTimeDiv.attr("aria-hidden", true);
      calendarTimeDiv.attr("tabindex", "-1");
      calendarTimeContainer.append(calendarTimeDiv);

      // Create Validators for the Virtual Fields
      addRequiredValidator("selectTime_date");
      addRequiredValidator("selectTime_time");
      configureRequiredLabels({
         fields: ["selectTime_date", "selectTime_time"]
      });

      // Render the Calendar
      try {
         calendarApi_Date = new FullCalendar.Calendar(calendarDateDiv[0], {
            // themeSystem: 'bootstrap',
            initialView: 'dayGridMonth',
            locale: getTranslatedText("en", "fr"),
            timeZone: 'UTC',
            firstDay: 0,
            showNonCurrentDates: true,
            fixedWeekCount: false,
            businessHours: true,
            nowIndicator: true,
            selectMirror: true,
            validRange: function (nowDate) {
               // Calculate the end date as either the last available day of the timeslots, or 6 months from now
               let end = "";
               let lastOption = $("#selectTime_date").find("option").last();
               let lastDate = lastOption.attr("data-date");

               if (!isNullOrWhitespace(lastDate)) {
                  end = moment.utc(lastDate).add(1, "day").format("YYYY-MM-DD");
               } else if (variableExists(selectTime_Office)) {
                  let dataKeys = Object.keys(selectTime_Office.AvailableWorkstations);
                  end = dataKeys[dataKeys.length - 1];
               }
               else { end = moment(nowDate).add(6, 'month').toISOString(); }

               return {
                  start: nowDate,
                  end: end
               };
            },
            events: function (info, successCallback, failureCallback) {
               if (!variableExists(selectTime_Office)) {
                  console.log("Office not selected. Events can't be retrieved");
                  successCallback([]);
                  return;
               }

               console.log("Fetching Calendar Events...");
               let calendarEvents = createFullCalendarDateEvents(selectTime_Office, {
                  displayTitle: false,
                  useUtc: true,
                  groupId: "",
                  selectedColor: "red",
                  deselectedColor: ""
               });
               successCallback(calendarEvents.events);
            },
            eventClick: function (info) {
               console.log('Event: ' + info.event.title);
               console.log('Coordinates: ' + info.jsEvent.pageX + ',' + info.jsEvent.pageY);
               console.log('View: ' + info.view.type);

               // Update the color on the Calendar Event
               let events = calendarApi_Date.getEvents();
               for (let i = 0; i < events.length; i++) {
                  let event = events[i];
                  event.setProp("backgroundColor", event.extendedProps.options.deselectedColor);
               }

               info.event.setProp("backgroundColor", info.event.extendedProps.options.selectedColor);

               // Update the Date Selector
               let selectTime_Date = $("#selectTime_date");
               selectTime_Date.val(info.event.extendedProps.dateKey);
               selectTime_Date.change();
            }
         });

         calendarApi_Time = new FullCalendar.Calendar(calendarTimeDiv[0], {
            // themeSystem: 'bootstrap',
            initialView: 'listDay', //'timeGridDay',
            locale: getTranslatedText("en", "fr"),
            timeZone: 'UTC',
            businessHours: true,
            nowIndicator: true,
            now: moment.utc().hours(new Date().getHours()).minutes(new Date().getMinutes()).seconds(new Date().getSeconds()).toISOString(),
            selectMirror: true,
            dayHeaders: false,
            allDaySlot: false,
            listDayFormat: false,
            listDaySideFormat: false,
            headerToolbar: {
               start: "title",
               center: "",
               end: ""
            },
            events: function (info, successCallback, failureCallback) {
               if (!variableExists(selectTime_Office)) {
                  console.log("Office not selected. Events can't be retrieved");
                  successCallback([]);
                  return;
               }

               console.log("Fetching Calendar Events...");
               let calendarEvents = createFullCalendarTimeEvents(selectTime_Office, {
                  displayTitle: true,
                  useUtc: true,
                  groupId: "",
                  selectedColor: "red",
                  deselectedColor: ""
               });
               successCallback(calendarEvents.events);
            },
            eventClick: function (info) {
               console.log('Event: ' + info.event.title);
               console.log('Coordinates: ' + info.jsEvent.pageX + ',' + info.jsEvent.pageY);
               console.log('View: ' + info.view.type);

               // Update the color on the Calendar Event
               let events = calendarApi_Time.getEvents();
               for (let i = 0; i < events.length; i++) {
                  let event = events[i];
                  event.setProp("backgroundColor", event.extendedProps.options.deselectedColor);
               }

               info.event.setProp("backgroundColor", info.event.extendedProps.options.selectedColor);

               // Update the Time Selector
               let selectTime_Time = $("#selectTime_time");
               selectTime_Time.val(info.event.extendedProps.dateKey);
               selectTime_Time.change();
            }
         })
      }
      catch (ex) {
         console.log("There was an error initializing Full-Calendar");
         console.log(ex);
      }

      // Create a container for actions
      let actionsContainer = $(document.createElement("div"));
      actionsContainer.attr("id", "step1_tab_section_selecttime_fieldset_actions");
      actionsContainer.addClass("pull-right");
      $(actionsContainer).insertAfter($("#step1_tab_section_selecttime_fieldset").find('legend'));

      // Create a container for messaging
      let messagingContainer = $(document.createElement("div"));
      messagingContainer.attr("id", "step1_tab_section_selecttime_fieldset_messaging");
      $(messagingContainer).insertAfter($("#step1_tab_section_selecttime_fieldset").find('legend'));
   });

   createInitializeLogic(initalizeLogic, "officeadditionalservice_subgrid", function () {
      let subgridId = this.jquery.attr("data-subgridId");
      let subgridData = configureSubgridUsability(subgridId, {
         isRequired: true,
         createDisableButton: true
      });

      $(subgridData.subgridData.primaryAction).parent().parent().removeClass("pull-right");

      let noBtn = createButton("Nobutton", "", {
         elementType: "button",
         buttonType: "btn-info",
         content: {
            text: getTranslatedText("No", "Non"),
         },
         position: "right",
         attributes: [
            { key: "data-subgridid", value: subgridId },
            { key: "aria-describedby", value: $("#officeadditionalservice_subgrid_create").attr("aria-describedby") },
         ],
         classes: [],
         css: [{ key: "margin-left", value: "5px" }],
         action: function () {
            let subgridId = $(this).attr("data-subgridid");
            let officeSelectionSection = $("#step1_tab_section_findoffice_fieldset");
            let additionalServiceSection = $("[data-name='step1_tab_section_additionalservices']");

            $("#" + subgridId).attr("data-lookup-viewed-once", true);
            sessionSaveObject(subgridId + "_viewonce", true);
            $("#" + subgridId + "_viewonce_error").hide();

            officeSelectionSection.show();
            additionalServiceSection.parents("details").removeAttr("open");
            // additionalServiceSection.parent().hide();

            // Remove all selections
            console.log("Removing all Additional Service Selections");
            $("#" + subgridId).find(".disassociate-link").click();
         }
      });

      $("#officeadditionalservice_subgrid_create").removeClass("pull-right");
      $("#officeadditionalservice_subgrid_create").removeAttr("title");
      $("#officeadditionalservice_subgrid_create_fake").parent().append(noBtn);

      // Adding Custom Event Handler to the Button
      $("#" + subgridId + "_create").click(function () {
         let subgridId = $(this).attr("data-subgridid");
         let officeSelectionSection = $("#step1_tab_section_findoffice_fieldset");

         $("#" + subgridId).attr("data-lookup-viewed-once", true);
         sessionSaveObject(subgridId + "_viewonce", true);

         officeSelectionSection.show();
         $("#" + subgridId + "_viewonce_error").hide();
      });

      createValidator(subgridId, null, "LookupViewedOnce", subgridId, false,
         getTranslatedText(
            "Please select an option to proceed with booking",
            "Veuillez sélectionner une option pour procéder à la réservation"
         ), function () {
            let subgridId = this.jquery.attr("data-subgridId");

            let lookupViewedOnce = $("#" + subgridId).attr("data-lookup-viewed-once");
            let sessionLookupViewedOnce = sessionLoadObject(subgridId + "_viewonce");

            return (lookupViewedOnce == true || sessionLookupViewedOnce == true);
         }.bind(this)
      );

      let sessionViewOnceValue = sessionLoadObject(subgridId + "_viewonce");
      let lookOnceAlert = createAlert(subgridId + "_viewonce_error", "alert-danger",
         getTranslatedText(
            'Before you continue, please select an option (Yes/No)',
            'Avant de continuer, veuillez sélectionner une option (Oui/Non)'
         ), (sessionViewOnceValue == false || sessionViewOnceValue == null),
         {
            backingId: "subgridId",
         }
      );

      // Create additionalServiceInfo Tooltips
      loadMetadata("/_api/esdc_officeadditionalservices?$filter=statecode eq 0", null, {}, {}).then(function (res) {
         let esdc_officeadditionalservices = res;

         let servicesDiv = $(document.createElement("div"));
         servicesDiv.attr("id", "additionalServiceInfo");
         servicesDiv.addClass("row");
         ($("[data-name='step1_tab_section_additionalservices']")).prepend(servicesDiv);
         if (variableExists(esdc_officeadditionalservices) && arrayExistsWithLength(esdc_officeadditionalservices.value, 0)) {
            let servicesTextHeader = $(document.createElement("p"));
            servicesTextHeader.addClass("h3");
            servicesTextHeader.css("text-align", "left");
            servicesTextHeader.text(getTranslatedText(
               'The following are examples of additional service needs you may require:',
               'Voici des exemples de services supplémentaires dont vous pourriez avoir besoin:'
            ));
            servicesDiv.append(servicesTextHeader);
            let servicesUl = $(document.createElement("ul"));
            servicesUl.attr("id", "optionsAdditionalServiceInfo");
            servicesDiv.append(servicesUl);
            servicesDiv.append(lookOnceAlert);
            try {
               for (let iService = 0; iService < esdc_officeadditionalservices.value.length; iService++) {
                  let service = esdc_officeadditionalservices.value[iService];
                  if (service.esdc_displayonportal == true) {
                     let serviceLi = $(document.createElement("li"));
                     serviceLi.attr("data-recordid", service.esdc_officeadditionalserviceid);
                     serviceLi.attr("data-name-english", service.esdc_nameenglish);
                     serviceLi.attr("data-name-french", service.esdc_namefrench);
                     serviceLi.text(getTranslatedText(
                        service.esdc_nameenglish,
                        service.esdc_namefrench
                     ));
                     servicesUl.append(serviceLi);
                  }
               }
            } catch (ex) { }
         }
      });

      // Create The Mask Exemption and Chemical Sensitiviies Validation
      createValidator(subgridId, null, "MaskExemption", subgridId, false,
         getTranslatedText(
            "Mask Exemption",
            "Exemption de masque"
         ), function () {
            let additionalServices = calculateBookingAdditionalServices({
               includeAssociatedServices: true,
               includeVirtualServices: false
            });

            let maskExemptionService = additionalServices.find(x => x.id.toLowerCase() == "b41478f1-2152-ec11-8f8e-000d3a0a00fb");
            if (maskExemptionService != undefined) {
               return false;
            }

            return true;
         }.bind(this), {
         error: {
            appendFieldNameToErrorMessage: false,
         }
      });

      // Create hidden input to cache saved selections
      let subgridInput = document.createElement("input");
      $(subgridInput).attr("id", subgridId + "_selectedservices");
      $(subgridInput).attr("type", "hidden");
      $(subgridInput).val("");
      $(subgridInput).insertBefore($("#" + subgridId));

      // Hook up the refocus
      $("#" + subgridId).find(".view-grid table").eq(0).attr("tabindex", "-1");
      $("#" + subgridId).find(".modal-associate").find("button").filter(".form-close,.primary,.cancel").click(function () {
         $("#" + subgridId).find(".view-grid table").eq(0).focus();
      });

      // Configure the Subgrid
      $("#" + subgridId).on("loaded", function (event) {
         console.log("officeadditionalservice_subgrid - Loaded");

         // Hide the columns depending on language
         let target = $(event.target);
         if (target.hasClass("entity-grid associate-lookup")) /* Lookup Modal */ {
            console.log("Associate Grid Loaded");
            let grid = $(this).find("table").eq(1);
            $(grid).removeClass("table-fluid"); // Remove the table-fluid class to fix the mobile view

            // Translate the existing Rows
            let headerRows = grid.find("thead tr");
            let bodyRows = grid.find("tbody tr");

            // Remove the () from the Headers and clear the content of the :before
            bodyRows.find("td").addClass("clear-content").attr("data-th", "");
            headerRows.find("th a").each(function () {
               let existingAriaLabel = $(this).attr("aria-label");
               let existingHtml = $(this).html();
               // console.log(`Existing:|${existingAriaLabel}|${existingText}|`);

               let newAriaLabel = existingAriaLabel.substring(0, existingAriaLabel.indexOf("(") - 1);
               let newHtml = existingHtml.slice(0, existingHtml.indexOf("(") - 1) + existingHtml.slice(existingHtml.indexOf(")") + 1);
               $(this).attr("aria-label", newAriaLabel);
               $(this).html(newHtml);
               // console.log(`New:|${newAriaLabel}|${newText}|`);
            });

            // Depending on if the page is English or French, set the required view
            if (isEnglish) {
               headerRows.each(function () { $(this).find("th").eq(2).hide(); });
               bodyRows.each(function () { $(this).find("td").eq(2).hide(); });
            }
            else {
               headerRows.each(function () { $(this).find("th").eq(1).hide(); });
               bodyRows.each(function () {
                  $(this).find("td").eq(1).hide();

                  // Microsoft didn't bother to implement Accessibility on non-english N2N Subgrids (◔_◔)
                  let checkboxes = $(this).find("td .fa.fa-fw");
                  checkboxes.attr("role", "checkbox");
                  checkboxes.attr("tabindex", 0);
                  checkboxes.attr("title", "Case à cocher");
                  checkboxes.attr("aria-label", "Sélectionner ou désélectionner la ligne");
               });
            }

            // Hide all rows that have already been added
            let additionalServices = calculateBookingAdditionalServices({ includeAssociatedServices: true, includeVirtualServices: true });
            $(bodyRows).each(function () {
               let rowId = $(this).attr("data-id");
               let foundService = additionalServices.find(x => {
                  return x.id.toLowerCase() == rowId.toLowerCase()
               });

               // Hide the row if the service is already selected
               if (foundService == undefined) { $(this).show(); }
               else { $(this).hide(); }
            });
         }
         else /* Subgrid */ {
            console.log("Subgrid Loaded");
            let grid = $(this).find("table").eq(0);
            grid.find("thead").addClass("sr-only");
            $(grid).removeClass("table-fluid"); // Remove the table-fluid class to fix the mobile view

            // Translate the existing Rows
            let headerRows = grid.find("thead tr");
            let bodyRows = grid.find("tbody tr");

            // Remove the () from the Headers and clear the content of the :before
            bodyRows.find("td").addClass("clear-content").attr("data-th", "");
            headerRows.find("th a").each(function () {
               let existingAriaLabel = $(this).attr("aria-label");
               let existingHtml = $(this).html();
               // console.log(`Existing:|${existingAriaLabel}|${existingText}|`);

               let newAriaLabel = existingAriaLabel.substring(0, existingAriaLabel.indexOf("(") - 1);
               let newHtml = existingHtml.slice(0, existingHtml.indexOf("(") - 1) + existingHtml.slice(existingHtml.indexOf(")") + 1);
               $(this).attr("aria-label", newAriaLabel);
               $(this).html(newHtml);
               // console.log(`New:|${newAriaLabel}|${newText}|`);
            });

            // Depending on if the page is English or French, set the required view
            let currentLanguageColumnIndex = 0;
            if (isEnglish) {
               currentLanguageColumnIndex = 0;
               headerRows.each(function () { $(this).find("th").eq(1).hide(); });
               bodyRows.each(function () {
                  $(this).find("td:eq(0),td:eq(1)").attr("tabindex", 0);
                  $(this).find("td").eq(1).hide(); // Hide the French Column

               });
            }
            else {
               currentLanguageColumnIndex = 1;
               headerRows.each(function () { $(this).find("th").eq(0).hide(); });
               bodyRows.each(function () {
                  $(this).find("td:eq(0),td:eq(1)").attr("tabindex", 0);
                  $(this).find("td").eq(0).hide();
               });
            }

            // Hide the View Once Error if we have multiple bodyRows because it implies they looked at the list
            if (bodyRows.length > 0) {
               $("#esdc_accessibilityinformation").parents("tr").show();
               $("#" + subgridId + "_viewonce_error").hide();
            }
            else { $("#esdc_accessibilityinformation").parents("tr").hide(); }

            // Enable special questions if an option is selected
            let allServices = sessionLoadObject("esdc_officeadditionalservices");
            let selectedServices = [];
            bodyRows.each(function () {
               let rowId = $(this).attr("data-id");
               let service = allServices.value.find(x => x.esdc_officeadditionalserviceid == rowId);
               if (service != undefined) {
                  selectedServices.push(service);
               }
            });

            // Loop selected services and enable sections if selected
            let selectedServiceIds = [];
            for (let i = 0; i < selectedServices.length; i++) {
               let iService = selectedServices[i];
               selectedServiceIds.push(iService.esdc_officeadditionalserviceid);

               let tdService = bodyRows.filter("[data-id='" + iService.esdc_officeadditionalserviceid + "']").children().eq(currentLanguageColumnIndex);
               $(tdService).html("<p>" + $(tdService).html() + "</p>");

               // Check for Mask Exemption
               if (iService.esdc_officeadditionalserviceid == "b41478f1-2152-ec11-8f8e-000d3a0a00fb") {
                  // Add in Error message for "Mask Exemption addtional service selected"
                  let cached_inputs = getCachedRecordInputs();
                  // BIOmetrics
                  if (cached_inputs.getValue("esdc_bookingtype", true) == "564190000") /* BIO */ {
                     let maskAlert = createAlert(subgridId + "_maskexemption_error", "alert-danger",
                        getTranslatedText(
                           "Biometrics Collections<br/>If you are requesting a Biometrics Collection service, you will need to call 1-833-283-7400 (TTY: 1-800-465-7735) to have your appointment booked.",
                           "Collecte biométrique<br/>Si vous demandez un service de collecte de données biométriques, vous devez appeler le 1-833-283-7400 (ATS : 1-800-465-7735) pour obtenir une aide avec la prise de rendez-vous."
                        ), true,
                        {
                           backingId: "subgridId",
                           containerAttributes: [
                              { key: "data-serviceid", value: "b41478f1-2152-ec11-8f8e-000d3a0a00fb" }
                           ]
                        }
                     );

                     $(tdService).append(maskAlert);
                  } else if (cached_inputs.getValue("esdc_bookingtype", true) == "564190001") /* PASSPORT */ {
                     let maskAlert = createAlert(subgridId + "_maskexemption_error", "alert-danger",
                        getTranslatedText(
                           "Passport Service<br/>If you are requesting a Passport Service, you will need to call 1-800-567-6868 (TTY: 1-866-255-7655) to have your appointment booked.",
                           "Services de passeport<br/>Si vous demandez un service de passeport, vous devez appeler le 1-800-567-6868 (ATS : 1-866-255-7655) pour obtenir une aide avec la prise de rendez-vous."
                        ), true,
                        {
                           backingId: "subgridId",
                           containerAttributes: [
                              { key: "data-serviceid", value: "b41478f1-2152-ec11-8f8e-000d3a0a00fb" }
                           ]
                        }
                     );

                     $(tdService).append(maskAlert);
                  }
               }

               // Check for Chemical Sensitivity
               if (iService.esdc_officeadditionalserviceid == "efd16ffd-2152-ec11-8f8e-000d3a0a00fb") {
                  //Create Information alert for Multiple Chemical Sensitivity
                  let chemicalAlert = createAlert(subgridId + "_chemicalsensitivity_info", "alert-info",
                     getTranslatedText(
                        "Service Canada promotes a scent-reduced environment. We encourage limited use of scented products in our offices.",
                        "Service Canada favorise un environnement de travail à parfum réduit. Nous encourageons les visiteurs à utiliser le moins possible les produits parfumés dans nos bureaux."
                     ), true,
                     {
                        backingId: "subgridId",
                        containerAttributes: [
                           { key: "data-serviceid", value: "efd16ffd-2152-ec11-8f8e-000d3a0a00fb" }
                        ]
                     }
                  );

                  $(tdService).append(chemicalAlert);

               }
            }

            // Save the Ids to a data attribute on the subgrid and hidden input for caching purposes
            $("#" + subgridId).attr("data-selected-service-ids", selectedServiceIds.join());
            $("#" + subgridId + "_selectedservices").val($("#" + subgridId).attr("data-selected-service-ids"));
            $("#" + subgridId + "_selectedservices").change()

            // Move the action links
            $(".disassociate-link").each((i, value) => {
               $(value).removeAttr("tabindex");
               $(value).css("text-decoration", "none");

               let td = $(value).parents("td").eq(0);
               td.children().hide();
               td.prepend(value)
            });
            $(".disassociate-link").find("span.fa").hide();
            $(".view-empty.message").hide();
         }
      });

      //Accessibility issue - Booking N2N "Selected Records" not translated on French Page
      if (!isEnglish) {
         $(".selected-records").on('DOMNodeInserted', function (e) {
            if ($(e.target).hasClass('item pull-left btn btn-default')) {
               let elementId = $(e.target).attr("data-id");
               let frText = $(".entity-grid").first().find("table").eq(1).find("tbody").find("[data-id='" + elementId + "']").find(`[data-attribute='esdc_namefrench']`).text();
               $(e.target).find("span.name").first().text(frText);
            }
         });
      }
   });

   createInitializeLogic(initalizeLogic, "esdc_officeid", function () {
      $("#esdc_officeid").parents("tr").hide();
      addRequiredValidator("esdc_officeid", null);
      configureRequiredLabels({
         fields: ["esdc_officeid"]
      });

      $("#esdc_officeid").change(function () {
         let lookup = getLookupField("esdc_officeid");

         // If we dont have an Office Id selected, then hide the section and exit
         if (lookup == null) { loadSelectTimeSection([]); return; }
         loadSelectTimeSection([lookup.id]);
      });
   });

   createInitializeLogic(initalizeLogic, "esdc_officeworkstationid", function () {
      $("#esdc_officeworkstationid").parents("tr").hide();
      addRequiredValidator("esdc_officeworkstationid", null);
      configureRequiredLabels({
         fields: ["esdc_officeworkstationid"]
      });
   });

   createInitializeLogic(initalizeLogic, "esdc_officeworkstationscheduleid", function () {
      $("#esdc_officeworkstationscheduleid").parents("tr").hide();
      addRequiredValidator("esdc_officeworkstationscheduleid", null);
      configureRequiredLabels({
         fields: ["esdc_officeworkstationscheduleid"]
      });
   });

   createInitializeLogic(initalizeLogic, "esdc_startdatetimeofficelocal", function () {
      $("#esdc_startdatetimeofficelocal").parents("tr").hide();
      addRequiredValidator("esdc_startdatetimeofficelocal", null);
      configureRequiredLabels({
         fields: ["esdc_startdatetimeofficelocal"]
      });

      $("#esdc_startdatetimeofficelocal_datepicker_description").removeAttr("aria-describedby");
   });

   createInitializeLogic(initalizeLogic, "esdc_enddatetimeofficelocal", function () {
      $("#esdc_enddatetimeofficelocal").parents("tr").hide();
      addRequiredValidator("esdc_enddatetimeofficelocal", null);
      configureRequiredLabels({
         fields: ["esdc_enddatetimeofficelocal"]
      });
   });

   createInitializeLogic(initalizeLogic, "esdc_applicationsubmittedon", function () {
      createValidator("esdc_applicationsubmittedon", null, "UpdateSubmittedOn", "esdc_applicationsubmittedon", false, "",
         function () {
            try {
               setDateTimeField("esdc_applicationsubmittedon", moment.utc().toDate(), true, false);
            } catch (ex) { }
            return true;
         });
   });
   createInitializeLogic(initalizeLogic, "esdc_clientdataentrycomplete", function () {
      createValidator("esdc_clientdataentrycomplete", null, "RequireYes", "esdc_clientdataentrycomplete", true, "",
         function () {
            setDateTimeField("esdc_applicationsubmittedon", moment.utc().toDate(), true, false);
            $("#esdc_clientdataentrycomplete").val("1");
            return true;
         }
      );
   });
   //#endregion

   return initalizeLogic;
}

/**
 * 
 * @param {string} dateISOShort The Date string formatted as YYYY-MM-DD 
 * @returns {[string]: any[]} The timeslots for this date, as a Dictionary using the time code
 */
function groupTimeslotsForDate(dateISOShort) {
   if (!variableExists(selectTime_OfficeAvailabilities)) { return {}; }
   if (!variableExists(selectTime_Office)) { return {}; }
   let requiredNumberTimeslots = calculateBookingDurationSlots();

   let selectedDay = selectTime_Office.AvailableWorkstations[dateISOShort];
   if (!variableExists(selectedDay)) {
      console.log("selectedDay is undefined for: " + dateISOShort);
      hideAndlockNextButton();
      return {};
   }

   // Convert Workstations and Timeslots to linear list for grouping purposes
   let availableTimeslots = [];
   for (let i = 0; i < selectedDay.length; i++) {
      let workstationTimeslots = selectedDay[i].AvailableTimeslots;
      for (let x = 0; x < workstationTimeslots.length; x++) {
         availableTimeslots.push(workstationTimeslots[x]);
      }
   }

   // Filter the linear list of timeslots by the required number of timeslots for this booking
   let selectedDateSlots = availableTimeslots.filter(x => {
      return (x.SequentialCount + 1) >= requiredNumberTimeslots;
   }).sort((a1, b1) => {
      let a = moment.utc(a1.CalendarFrom);
      let b = moment.utc(b1.CalendarFrom);
      return (a.unix() > b.unix()) ? 1 : ((b.unix() > a.unix()) ? -1 : 0);
   });

   // Group the remaining timeslots into a dynamic object using the time key
   let groupedTimeslots = groupArrayByKey(selectedDateSlots, "CalendarFrom", {
      key: function (key) {
         return moment.utc(key).format("HH:mm");
      }
   });

   // Return the Timeslots grouped by the From Time
   return groupedTimeslots
}

/**
 * Generates events for the Full Calendar implementation
 * @param {any} selectedOffice The Selected Office the events are on
 * @param {{
 *    displayTitle: boolean,
 *    useUtc: boolean,
 *    groupId: string,
 *    selectedColor: string,
 *    deselectedColor: string,
 * }} options Additional rendering options for the events
 * @returns {{
 *    businessHours: any[],
 *    events: any[]
 * }} Array of Full Calendar Events
 */
function createFullCalendarDateEvents(selectedOffice, options) {
   let dataKeys = Object.keys(selectedOffice.AvailableWorkstations);

   // Calculate the Business Hours
   let businessHours = [];

   // Transform the Events
   let events = dataKeys.map((x, index) => {
      let id = x + "_" + index;
      let title = options.displayTitle ? x : "";
      let group = isNullOrWhitespace(options.groupId) ? "" : options.groupId; // "office-month-hoursofoperation"
      let start = options.useUtc ? moment.utc(x).toISOString() : moment(x).toISOString();
      let end = options.useUtc ? moment.utc(x).toISOString() : moment(x).toISOString();
      return {
         id: id,
         groupId: group,
         title: title,
         start: start,
         end: end,
         display: "background",
         allDay: true,
         extendedProps: {
            office: selectedOffice.Office,
            options: options,
            dateKey: x,
            index: index,
         }
      }
   });

   return {
      businessHours: businessHours,
      events: events
   };
}

/**
 * Generates events for the Full Calendar implementation
 * @param {any} selectedOffice The Selected Office the events are on
 * @param {{
 *    displayTitle: boolean,
 *    useUtc: boolean,
 *    groupId: string,
 *    selectedColor: string,
 *    deselectedColor: string,
 * }} options Additional rendering options for the events
 * @returns {{
 *    businessHours: any[],
 *    events: any[]
 * }} Array of Full Calendar Events
 */
function createFullCalendarTimeEvents(selectedOffice, options) {
   let dataKeys = Object.keys(selectedOffice.AvailableWorkstations);
   let timeOptions = $("#selectTime_time").find("option:not([value=-1])");

   // Calculate the Business Hours
   let businessHours = [];

   // Transform the Events
   // index and x flipped because jquery map is different order from ES5 Map
   let events = timeOptions.map((index, x) => {
      let startIso = $(x).attr("data-datetime-iso");
      let start12hour = $(x).attr("data-time-12hour");
      let start24hour = $(x).attr("data-time-24hour");
      let text = $(x).text();
      let value = $(x).attr("value");

      let id = value + "_" + index;
      let title = options.displayTitle ? getTranslatedText(start24hour, start12hour) : "";
      let group = isNullOrWhitespace(options.groupId) ? "" : options.groupId; // "office-month-hoursofoperation"
      let start = startIso;
      return {
         id: id,
         groupId: group,
         title: title,
         start: start,
         allDay: false,
         extendedProps: {
            office: selectedOffice.Office,
            options: options,
            dateKey: value,
            timeIso: startIso,
            start12hour: start12hour,
            time24Hour: start24hour,
            optionText: text,
            index: index,
         }
      }
   });

   return {
      businessHours: businessHours,
      events: events
   };
}

/**
 * Generates events for the WET Calendar
 * @param {any} selectedOffice The selected Office the events are on
 * @returns {any[]} Array of WET Calendar Events
 */
function createWETEvents(selectedOffice) {
   let dataKeys = Object.keys(selectedOffice.AvailableWorkstations);
   let events = dataKeys.map(x => {
      return {
         title: {
            text: x,
            href: "#",
            onClick: function (event) {
               event.preventDefault();

               let date = $(this).attr("data-date");
               let officeId = $(this).attr("data-office-id");
               let backingId = $(this).attr("data-backing-id");
               console.log(`Calendar Event Clicked: ${officeId}(${date})`);

               $("#" + backingId).val(date);
               $("#" + backingId).change();
            },
            attributes: [
               { key: "data-date", value: x },
               { key: "data-office-id", value: selectedOffice.Office.id },
               { key: "data-backing-id", value: "selectTime_date" }
            ]
         },
         time: new Date(x),
         enableListFiltering: true
      }
   });

   return events;
}

/**
 * Creates the Mapping Search Functionality with browser Support
 * @param {string} id The ID of the Search Field
 * @param {{
 *    displayLimit: {
 *       visible: boolean,
 *       optionMode: "options",
 *       defaultValue: number,
 *       widthPx: number,
 *       options?: {
 *          value: number,
 *          text: string
 *       }[]
 *    }
 *    distance: {
 *       visible: boolean,
 *       optionMode: "options"|"calculate",
 *       defaultValue: number,
 *       widthPx: number,
 *       options?: {
 *          distance: number,
 *          zoomLevel: number
 *       }[],
 *       calculate?: {
 *          minKm: number,
 *          maxKm: number,
 *          stepKm: number,
 *          zoomLevel: number
 *       }      
 *    },
 *    mapSettings: {
 *       height: string
 *    }
 * }} options The Mapping Options
 * @returns {HTMLDivElement} The Mapping container
 */
function createMappingStructured(id, options) {
   if (!variableExists(options.distance)) { options.distance = {}; }
   if (!variableExists(options.distance.widthPx)) { options.distance.widthPx = 100; }
   if (!variableExists(options.distance.visible)) { options.distance.visible = true; }
   if (!variableExists(options.distance.options)) {
      options.distance.options = [
         { distance: 25, zoomLevel: 11 },
         { distance: 50, zoomLevel: 11 },
         { distance: 100, zoomLevel: 11 },
         { distance: 150, zoomLevel: 11 },
         { distance: 200, zoomLevel: 11 },
         { distance: 250, zoomLevel: 11 },
      ];
   }
   if (!variableExists(options.distance.calculate)) { options.distance.calculate = {}; }
   if (!variableExists(options.distance.calculate.minKm)) { options.distance.calculate.minKm = 50; }
   if (!variableExists(options.distance.calculate.maxKm)) { options.distance.calculate.maxKm = 300; }
   if (!variableExists(options.distance.calculate.stepKm)) { options.distance.calculate.stepKm = 50; }
   if (!variableExists(options.distance.calculate.zoomLevel)) { options.distance.calculate.zoomLevel = 11; }
   if (isNullOrWhitespace(options.distance.optionMode)) { options.distance.optionMode = "calculate"; }
   if (!variableExists(options.distance.defaultValue)) { options.distance.defaultValue = 50; }

   // Create the Parent Container
   let azureMapsContainer = $(document.createElement("div"));
   azureMapsContainer.attr("id", id);
   azureMapsContainer.attr("data-mapid", id + "_map");
   azureMapsContainer.attr("data-distance-id", id + "_distance");
   azureMapsContainer.attr("data-displaylimit-id", id + "_displaylimit");
   azureMapsContainer.attr("data-search-progressid", id + "_search_loading");
   azureMapsContainer.attr("data-button-requestlocationid", id + "_RequestLocationButton");
   azureMapsContainer.attr("data-button-searchid", id + "_MapSearchButton");
   // azureMapsContainer.addClass("row");

   // Create the Structured Search
   let searchContainer = $(document.createElement("div"));
   searchContainer.attr("id", id + "_search");
   searchContainer.addClass("row");
   azureMapsContainer.append(searchContainer);

   let searchColumnClasses = "col-xs-12 col-sm-12 col-md-6 col-lg-3";

   // Search Header
   let searchHeader = $(document.createElement("h3"));
   searchHeader.addClass("col-xs-12");
   searchHeader.text(getTranslatedText("Address Search", "Recherche d'adresse"));
   searchContainer.append(searchHeader);

   // Address
   let addressInput = createPortalField(id + "_search_address", {
      labelText: getTranslatedText("Address", "Adresse")
   }, false, {
      containerType: "div",
      elementType: "input",
      attributes: [
         { key: "type", value: "text" },
         { key: "placeholder", value: getTranslatedText("", "") },
         { key: "autocomplete", value: "street-address" },
         { key: "data-mapid", value: id + "_map" },
         { key: "data-distance-id", value: id + "_distance" },
         { key: "data-displaylimit-id", value: id + "_displaylimit" },
         { key: "data-button-requestlocationid", value: id + "_RequestLocationButton" },
         { key: "data-button-searchid", value: id + "_MapSearchButton" },
      ],
      handlers: [
         {
            eventName: "keydown",
            handler: function (e) {
               let searchButton = $(this).attr("data-button-searchid");
               if (e.ctrlKey && e.keyCode == 13) /* Ctrl-Enter */ {
                  $("#" + searchButton).click();
               }
               else if (e.keyCode == 13) /* Enter */ {
                  $("#" + searchButton).click();
               }
            }
         }
      ]
   });
   $(addressInput).addClass(searchColumnClasses);
   searchContainer.append(addressInput);

   // City
   let cityInput = createPortalField(id + "_search_city", {
      labelText: getTranslatedText("City", "Ville")
   }, false, {
      containerType: "div",
      elementType: "input",
      attributes: [
         { key: "type", value: "text" },
         { key: "placeholder", value: getTranslatedText("", "") },
         { key: "autocomplete", value: "address-level2" },
         { key: "data-mapid", value: id + "_map" },
         { key: "data-distance-id", value: id + "_distance" },
         { key: "data-displaylimit-id", value: id + "_displaylimit" },
         { key: "data-button-requestlocationid", value: id + "_RequestLocationButton" },
         { key: "data-button-searchid", value: id + "_MapSearchButton" },
      ],
      handlers: [
         {
            eventName: "keydown",
            handler: function (e) {
               let searchButton = $(this).attr("data-button-searchid");
               if (e.ctrlKey && e.keyCode == 13) /* Ctrl-Enter */ {
                  $("#" + searchButton).click();
               }
               else if (e.keyCode == 13) /* Enter */ {
                  $("#" + searchButton).click();
               }
            }
         }
      ]
   });
   $(cityInput).addClass(searchColumnClasses);
   searchContainer.append(cityInput);

   // Province/Territory/State
   let provinceTerritorySelect = createPortalField(id + "_search_provinceterritory", {
      labelText: getTranslatedText("Province/Territory", "Province / Territoire")
   }, false, {
      containerType: "div",
      elementType: "select",
      options: createOptionList(sessionLoadObject("esdc_provinceterritorystates").value.filter(function (x) {
         if (x._esdc_countryid_value == "0cf5389e-97ae-eb11-8236-000d3af4bfc3") /* Canada */ {
            return true;
         }
         return false;
      }).map(function (x) {
         return {
            text: getTranslatedText(x.esdc_nameenglish, x.esdc_namefrench),
            value: x.esdc_provinceterritorystateid,
            groupId: x.esdc_groupkey,
            attributes: [
               { key: "data-name-english", value: x.esdc_nameenglish },
               { key: "data-name-french", value: x.esdc_namefrench },
               { key: "data-countryid", value: x._esdc_countryid_value }
            ]
         }
      }), { text: getTranslatedText("Select your Province/Territory", "Sélectionnez votre province/territoire") }).map(x => { return { element: x }; }),
      attributes: [
         { key: "data-mapid", value: id + "_map" },
         { key: "data-distance-id", value: id + "_distance" },
         { key: "data-displaylimit-id", value: id + "_displaylimit" },
         { key: "data-button-requestlocationid", value: id + "_RequestLocationButton" },
         { key: "data-button-searchid", value: id + "_MapSearchButton" },
      ],
      handlers: [
         {
            eventName: "keydown",
            handler: function (e) {
               let searchButton = $(this).attr("data-button-searchid");
               if (e.ctrlKey && e.keyCode == 13) /* Ctrl-Enter */ {
                  $("#" + searchButton).click();
               }
               else if (e.keyCode == 13) /* Enter */ {
                  $("#" + searchButton).click();
               }
            }
         }
      ]
   });
   $(provinceTerritorySelect).addClass(searchColumnClasses);
   searchContainer.append(provinceTerritorySelect);

   // Postal Code
   let postalCodeInput = createPortalField(id + "_search_postalcode", {
      labelText: getTranslatedText("Postal Code", "Code postal")
   }, false, {
      containerType: "div",
      elementType: "input",
      attributes: [
         { key: "type", value: "text" },
         { key: "placeholder", value: getTranslatedText("", "") },
         { key: "autocomplete", value: "postal-code" },
         { key: "data-mapid", value: id + "_map" },
         { key: "data-distance-id", value: id + "_distance" },
         { key: "data-displaylimit-id", value: id + "_displaylimit" },
         { key: "data-button-requestlocationid", value: id + "_RequestLocationButton" },
         { key: "data-button-searchid", value: id + "_MapSearchButton" },
      ],
      handlers: [
         {
            eventName: "keydown",
            handler: function (e) {
               let searchButton = $(this).attr("data-button-searchid");
               if (e.ctrlKey && e.keyCode == 13) /* Ctrl-Enter */ {
                  $("#" + searchButton).click();
               }
               else if (e.keyCode == 13) /* Enter */ {
                  $("#" + searchButton).click();
               }
            }
         }
      ]
   });
   $(postalCodeInput).addClass(searchColumnClasses);
   searchContainer.append(postalCodeInput);

   // Create the Search Action Bar
   azureMapsContainer.append($("<hr>"));

   let actionContainer = $(document.createElement("div"));
   actionContainer.attr("id", id + "_actions");
   actionContainer.addClass("row");
   azureMapsContainer.append(actionContainer);

   // Distance
   let distanceDiv = createPortalField(id + "_distance", {
      labelText: getTranslatedText("Distance (Radius)", "Distance (rayon)")
   }, false, {
      containerType: "div",
      elementType: "select",
      attributes: [
         { key: "data-zoomlevel", value: 11 },
         { key: "data-mapid", value: id + "_map" },
         { key: "data-distance-id", value: id + "_distance" },
         { key: "data-displaylimit-id", value: id + "_displaylimit" },
         { key: "data-button-requestlocationid", value: id + "_RequestLocationButton" },
         { key: "data-button-searchid", value: id + "_MapSearchButton" },
      ],
      handlers: [
         {
            eventName: "change",
            handler: function () {
               let selectedValue = $(this).val();
               let option = $(this).find(`option[value="${selectedValue}"]`);
               $(this).attr("data-zoomlevel", $(option).attr("data-zoomlevel"));
            }
         },
         {
            eventName: "keydown",
            handler: function (e) {
               let searchButton = $(this).attr("data-button-searchid");
               if (e.ctrlKey && e.keyCode == 13) /* Ctrl-Enter */ {
                  $("#" + searchButton).click();
               }
               else if (e.keyCode == 13) /* Enter */ {
                  $("#" + searchButton).click();
               }
            }
         }
      ]
   });

   let distanceDiv_select = $(distanceDiv).find("select");
   if (options.distance.optionMode == "calculate") {
      for (let i = options.distance.calculate.minKm; i <= options.distance.calculate.maxKm; i += options.distance.calculate.stepKm) {
         $(distanceDiv_select).append(`<option value="${i}" data-zoomlevel="${options.distance.calculate.zoomLevel}">${i}km</option>`);
      }
   }
   else if (options.distance.optionMode == "options") {
      for (let i = 0; i < options.distance.options.length; i++) {
         let option = options.distance.options[i];
         $(distanceDiv_select).append(`<option value="${option.distance}" data-zoomlevel="${option.zoomLevel}">${option.distance}km</option>`);
      }
   }
   $(distanceDiv_select).val(options.distance.defaultValue);
   $(distanceDiv_select).change();

   $(distanceDiv).addClass("col-xs-12 col-sm-3 col-md-3 col-lg-3");
   if (!options.distance.visible) { $(distanceDiv).hide(); }
   actionContainer.append(distanceDiv);

   // Display Limit
   let displayLimitDiv = createPortalField(id + "_displaylimit", {
      labelText: getTranslatedText("Number of Offices Displayed", "Nombre de bureaux affichés")
   }, false, {
      containerType: "div",
      elementType: "select",
      attributes: [
         { key: "data-mapid", value: id + "_map" },
         { key: "data-distance-id", value: id + "_distance" },
         { key: "data-displaylimit-id", value: id + "_displaylimit" },
         { key: "data-button-requestlocationid", value: id + "_RequestLocationButton" },
         { key: "data-button-searchid", value: id + "_MapSearchButton" },
      ],
      handlers: [
         {
            eventName: "change",
            handler: function () {
               let searchButton = $(this).attr("data-button-searchid");
               $("#" + searchButton).click();
            }
         },
         {
            eventName: "keydown",
            handler: function (e) {
               let searchButton = $(this).attr("data-button-searchid");
               if (e.ctrlKey && e.keyCode == 13) /* Ctrl-Enter */ {
                  $("#" + searchButton).click();
               }
               else if (e.keyCode == 13) /* Enter */ {
                  $("#" + searchButton).click();
               }
            }
         }
      ]
   });

   let displayLimitDiv_select = $(displayLimitDiv).find("select");
   if (options.displayLimit.optionMode == "options") {
      for (let i = 0; i < options.displayLimit.options.length; i++) {
         let option = options.displayLimit.options[i];
         $(displayLimitDiv_select).append(`<option value="${option.value}">${option.text}</option>`);
      }
   }

   $(displayLimitDiv_select).val(options.displayLimit.defaultValue);
   $(displayLimitDiv_select).change();

   $(displayLimitDiv).addClass("col-xs-12 col-sm-3 col-md-3 col-lg-3");
   if (!options.displayLimit.visible) { $(displayLimitDiv).hide(); }
   actionContainer.append(displayLimitDiv);

   // Button Containers
   let actionButtons = $(document.createElement("div"));
   actionButtons.addClass("col-xs-12 col-sm-9 col-md-9 col-lg-9");
   actionContainer.append(actionButtons);

   let actionPullRightDiv = $(document.createElement("div"));
   actionPullRightDiv.addClass("pull-right");
   actionButtons.append(actionPullRightDiv);

   // Location Button
   let locationButton = createButton(id + "_RequestLocationButton", "", {
      content: {
         text: getTranslatedText("Use my Location", "Utiliser mon emplacement"),
         glyphicon: {
            position: "left",
            type: "glyphicon-map-marker",
            aria: getTranslatedText("Use my Location", "Utiliser mon emplacement"),
            role: "button"
         }
      },
      buttonType: "btn-default",
      position: "left",
      attributes: [
         { key: "id", value: id + "_RequestLocationButton" },
         { key: "title", value: getTranslatedText("Use Location", "Utiliser l'emplacement") },
         { key: "data-prompted", value: false },
         { key: "data-latitude", value: "" },
         { key: "data-longitude", value: "" },
         { key: "data-accuracy", value: "" },
         { key: "data-backingId", value: id },
         { key: "data-mapid", value: id + "_map" },
         { key: "data-button-searchid", value: id + "_MapSearchButton" },
         { key: "data-map-enabled", value: variableExists(options.mapSettings) }
      ],
      action: function () {
         let backingId = $(this).attr("data-backingId");
         let mapId = $(this).attr("data-mapid");
         $(this).attr("data-prompted", true);
         $("#" + backingId).attr("data-prompted", true);

         let searchButtonId = $("#" + backingId).attr("data-button-searchid");
         let searchProgressId = $("#" + backingId).attr("data-search-progressid");
         $("#" + searchProgressId).show();

         try {
            queryGeolocationPermission().then(async (result) => {
               $("#" + backingId).trigger("onPermissionsChanged", [{
                  name: "geolocation",
                  result: result
               }]);

               // If this was iOS and we did the position call already, then just return that position
               if (variableExists(result.position)) { return result.position; }

               // If the result was denied, then exit early
               if (result.state == "denied") {
                  throw {
                     name: "geolocation",
                     state: result.state,
                     error: new Error("Permission has not been granted")
                  };
               }

               // Otherwise, go retrieve the position
               return await queryGeolocationPosition({
                  maximumAge: 0,
                  timeout: Infinity,
                  enableHighAccuracy: false
               });
            }).then((position) => {
               if (position == null) {
                  throw {
                     name: "geolocation",
                     state: "denied",
                     error: new Error("Permission was denied or disabled by browser")
                  };
               }

               $("#" + backingId).trigger("onPermissionsChanged", [{
                  name: "geolocation",
                  result: { state: "granted" }
               }]);

               sessionSaveObject("browser-location-position", position);
               let crd = position.coords;
               let searchQuery = `${crd.longitude},${crd.latitude}`;

               console.log('Your current position is:');
               console.log(`Latitude : ${crd.latitude}`);
               console.log(`Longitude: ${crd.longitude}`);
               console.log(`More or less ${crd.accuracy} meters.`);

               $(this).attr("data-latitude", crd.latitude);
               $(this).attr("data-longitude", crd.longitude);
               $(this).attr("data-accuracy", crd.accuracy);

               let backingField = $("#" + backingId);

               backingField.val(searchQuery);
               backingField.attr("data-searchquery-position", searchQuery);
               backingField.attr("data-browser-latitude", crd.latitude);
               backingField.attr("data-browser-longitude", crd.longitude);
               backingField.attr("data-browser-accuracy", crd.accuracy);

               try {
                  backingField.trigger("onPositionChange", [{
                     type: "GPS",
                     position: position,
                     latitude: crd.latitude,
                     longitude: crd.longitude,
                     accuracy: crd.accuracy,
                     searchQuery: searchQuery,
                     eventOrigin: {
                        backingField: backingField,
                        mapId: mapId,
                        map: $("#" + mapId),
                        button: $(this)
                     }
                  }]);
               } catch (ex) { console.log("There was an exception triggering the onPositionChange event"); }

               // backingField.removeAttr("readonly");
               return position;
            }).catch(ex => {
               console.log("Geolocation has been disabled by browser/organization settings");
               console.log(ex);

               let state = "denied";
               try {
                  if (ex["state"] != undefined) { state = ex["state"]; }
                  else if (ex instanceof GeolocationPositionError) { state = "denied"; }
               } catch (ex) { state = "denied"; }

               $("#" + backingId).trigger("onPermissionsChanged", [{
                  name: "geolocation",
                  result: { state: state }
               }]);

               // If we dont have permissions, redirect the request back into the Search Button
               $("#" + searchButtonId).click();
               return null;
            });
         }
         catch (ex) {
            console.log("There was an Exception requesting Location");
            console.log(ex);

            $("#" + backingId).trigger("onPermissionsChanged", [{
               name: "geolocation",
               result: { state: "denied" }
            }]);

            // If we dont have permissions, redirect the request back into the Search Button
            $("#" + searchButtonId).click();
         }
      }
   });
   actionPullRightDiv.append(locationButton);

   // Search Button
   let searchButton = createButton(id + "_MapSearchButton", "", {
      content: {
         text: getTranslatedText("Search", "Rechercher"),
         glyphicon: {
            position: "left",
            type: "glyphicon-search",
            aria: getTranslatedText("Search", "Rechercher"),
            role: "button"
         }
      },
      buttonType: "btn-default",
      position: "right",
      attributes: [
         { key: "id", value: id + "_MapSearchButton" },
         { key: "title", value: getTranslatedText("Search", "Rechercher") },
         { key: "data-backingId", value: id },
         { key: "data-search-fieldids", value: "_search_address,_search_city,_search_provinceterritory,_search_postalcode" },
         { key: "data-mapid", value: id + "_map" },
         { key: "data-button-locationid", value: id + "_RequestLocationButton" },
         { key: "data-map-enabled", value: variableExists(options.mapSettings) }
      ],
      action: async function () {
         let backingId = $(this).attr("data-backingId");
         let mapId = $(this).attr("data-mapid");

         let searchProgressId = $("#" + backingId).attr("data-search-progressid");
         $("#" + searchProgressId).show();

         // if (loadingOfficeFetch) {
         //    console.log("Office/FetchXml is currently being retrieved...");
         //    await waitUntil(() => { return loadingOfficeFetch == false; })
         //       .then(() => { console.log("Office/FetchXml retrieved. Continuing search"); })
         //       .catch(() => { console.log("An error occured with the office retrieval. Please try again. "); });
         // }

         let searchFieldIdsRaw = $(this).attr("data-search-fieldids");
         let requestLocationButtonId = $("#" + backingId).attr("data-button-requestlocationid");

         // Calculate the Search Query using the searchFieldIds
         if (isNullOrWhitespace(searchFieldIdsRaw)) { console.log("Please specify fields for search"); $("#" + searchProgressId).hide(); return; }
         let searchFieldIds = searchFieldIdsRaw.split(",");
         let values = [];
         for (let i = 0; i < searchFieldIds.length; i++) {
            let searchId = "#" + backingId + searchFieldIds[i];
            let field = $(searchId);

            let selectedValue = field.val();
            if (isNullOrWhitespace(selectedValue)) { continue; }

            if (field.is("select")) {
               if (selectedValue == "-1") { continue; }
               let parsedOption = field.find(`option[value=${selectedValue}]`);
               values.push(parsedOption.text());
            }
            else { values.push(selectedValue); }
         }

         let searchQuery = values.join(" ");
         if (isNullOrWhitespace(searchQuery)) { console.log("Please specify a search query"); $("#" + searchProgressId).hide(); return; }

         // Depending on if it was previously prompted, either invoke the search or redirect into location
         let wasPromptedPreviously = $("#" + requestLocationButtonId).attr("data-prompted");
         if (wasPromptedPreviously == "true") {
            $("#" + backingId).trigger("onSearch", [{
               searchQuery: searchQuery,
               mapId: mapId
            }]);
         }
         else {
            $("#" + requestLocationButtonId).click();
         }
      }
   });
   actionPullRightDiv.append(searchButton);

   // Create the Mapping and Progress Indicator
   azureMapsContainer.append($("<hr>"));

   // Create a Progress Indicator to display on the page
   let progressIndicator = createProgressIndicator(id + "_search_loading", {
      type: "circle",
      indeterminate: true,
      enabled: true,
      container: {
         position: "relative",
         class: "row",
         text: getTranslatedText(
            "Please wait while your search completes...",
            "Veuillez patienter pendant que nous effectuons votre recherche..."
         )
      },
      style: {
         display: "inline",
         position: "relative",
         left: "0px"
      },
      circle: {
         width: 100,
         height: 100,
         radius: 40
      }
   });
   $(progressIndicator).hide();
   azureMapsContainer.append(progressIndicator);

   let mapContainer = $(document.createElement("div"));
   mapContainer.attr("id", id + "_mapping");
   mapContainer.addClass("row");
   azureMapsContainer.append(mapContainer);

   let mapDiv = $(document.createElement("div"));
   mapDiv.attr("id", id + "_map");
   if (variableExists(options.mapSettings)) {
      mapDiv.css("width", "auto");
      mapDiv.css("height", options.mapSettings.height);
   }

   // Added an Event handler to initialze Azure Maps
   mapDiv.on("onInitializeAzureMaps", function (event, data) {
      let mapEnabled = $("#azureMapsMapEnabled").val();
      if (mapEnabled != "True") { console.log("Azure Maps is disabled"); return; }

      let mapDiv = $(this)[0];

      // Configure the Map
      if (mapService == null) {
         console.log("Configuring Map");
         mapService = InstantiateMappingService(mapDiv, {
            center: [data.longitude, data.latitude],
            interactive: $("#azureMapsInteractive").val() == "True",
            zoom: data.zoomLevel
         }, {});
         console.log("Map Initialized", mapService);
      }

      // Configure the Search
      if (searchService == null) {
         console.log("Configuring Search");
         searchService = InitializeSearchService("anonymous", mapService.map);
         console.log("Search Initialized", searchService);
      }
   });

   mapContainer.append(mapDiv);

   // Check the Geolocation Permissions to Pre-enable if permission has already been prompted
   // If so, toggle the Strucutred Search Button
   $(azureMapsContainer).on("onPermissionsChanged", function (event, data) {
      console.log("Permissions State has changed");
      let backingId = $(this).attr("id");
      let searchButtonId = $(this).attr("data-button-searchid");
      let requestLocationButtonId = $(this).attr("data-button-requestlocationid");
      let searchButton = $("#" + searchButtonId);
      let locationButton = $("#" + requestLocationButtonId);

      let name = data.name;
      let result = data.result;
      let state = data.result.state;
      console.log(`${name} - ${state}`);

      // Update the state
      $(searchButton).attr("data-geolocation-state", state);
      $(locationButton).attr("data-geolocation-state", state);
      $(this).attr("data-geolocation-state", state);

      // Reset the buttons to btn-default to ensue styling works later on
      $(locationButton).addClass("btn-default").removeClass("btn-primary").removeClass("btn-secondary").removeClass("btn-danger");
      $(searchButton).addClass("btn-default").removeClass("btn-primary").removeClass("btn-secondary").removeClass("btn-danger");

      // Update the views and set the proper flags depending on the state
      if (result.state == 'prompt') {
         $(searchButton).attr("data-prompted", false);
         $(locationButton).attr("data-prompted", false);
         $(this).attr("data-prompted", false);

         $(locationButton).removeAttr("disabled");
         $(searchButton).attr("disabled");
      }
      else if (result.state == "granted") {
         $(locationButton).addClass("btn-primary").removeClass("btn-default");
         $(searchButton).addClass("btn-secondary").removeClass("btn-default");

         $(searchButton).attr("data-prompted", true);
         $(locationButton).attr("data-prompted", true);
         $(this).attr("data-prompted", true);

         $(locationButton).removeAttr("disabled");
         $(searchButton).removeAttr("disabled");
      }
      else if (result.state == "denied") {
         $(locationButton).addClass("btn-danger").removeClass("btn-default");
         $(searchButton).addClass("btn-primary").removeClass("btn-default");

         $(searchButton).attr("data-prompted", true);
         $(locationButton).attr("data-prompted", true);
         $(this).attr("data-prompted", true);

         $(locationButton).attr("disabled", "disabled");
         $(searchButton).removeAttr("disabled");
      }
   });

   // if (!isSafari()) {
   queryGeolocationPermission(function (event) {
      console.log("queryGeolocationPermission - OnChangeCallback Triggered");
      $("#mappingSearch").trigger("onPermissionsChanged", [{
         name: "geolocation",
         result: this
      }]);
   }).then((result) => {
      $(azureMapsContainer).trigger("onPermissionsChanged", [{
         name: "geolocation",
         result: result
      }]);

      if (result.state == 'prompt') { return result; }
      console.log("Geolocation has already been granted/denied.");
      return result;
   }).catch(x => {
      console.log("Geolocation error has occured. Default: Denied.");
      console.log(x.error);
      $(azureMapsContainer).trigger("onPermissionsChanged", [{
         name: "geolocation",
         result: result
      }]);
      return x;
   });
   // }

   // Return the container
   return azureMapsContainer;
}

/**
 * Creates the Mapping Search Functionality with browser Support
 * @param {string} id The ID of the Search Field
 * @param {{
 *    distance: {
 *       optionMode: "options"|"calculate",
 *       defaultValue: number,
 *       widthPx: number,
 *       options?: {
 *          distance: number,
 *          zoomLevel: number
 *       }[],
 *       calculate?: {
 *          minKm: number,
 *          maxKm: number,
 *          stepKm: number,
 *          zoomLevel: number
 *       }      
 *    },
 *    mapSettings: {
 *       height: string
 *    }
 * }} options The Mapping Options
 * @returns {HTMLDivElement} The Mapping container
 */
function createMappingFuzzy(id, options) {
   if (!variableExists(options.distance)) { options.distance = {}; }
   if (!variableExists(options.distance.widthPx)) { options.distance.widthPx = 100; }
   if (!variableExists(options.distance.options)) {
      options.distance.options = [
         { distance: 25, zoomLevel: 11 },
         { distance: 50, zoomLevel: 11 },
         { distance: 100, zoomLevel: 11 },
         { distance: 150, zoomLevel: 11 },
         { distance: 200, zoomLevel: 11 },
         { distance: 250, zoomLevel: 11 },
      ];
   }
   if (!variableExists(options.distance.calculate)) { options.distance.calculate = {}; }
   if (!variableExists(options.distance.calculate.minKm)) { options.distance.calculate.minKm = 50; }
   if (!variableExists(options.distance.calculate.maxKm)) { options.distance.calculate.maxKm = 300; }
   if (!variableExists(options.distance.calculate.stepKm)) { options.distance.calculate.stepKm = 50; }
   if (!variableExists(options.distance.calculate.zoomLevel)) { options.distance.calculate.zoomLevel = 11; }
   if (isNullOrWhitespace(options.distance.optionMode)) { options.distance.optionMode = "calculate"; }
   if (!variableExists(options.distance.defaultValue)) { options.distance.defaultValue = 50; }

   let mappingSearchInputDiv = createPortalField(id,
      {
         labelText: "Search"
      }, true,
      {
         elementType: "input",
         readonly: true,
         attributes: [
            { key: "type", value: "text" },
            { key: "placeholder", value: getTranslatedText("Search", "Rechercher") },
            { key: "data-mapid", value: id + "_map" },
            { key: "data-distance-id", value: id + "_distance" },
            { key: "data-button-requestlocationid", value: id + "_RequestLocationButton" },
            { key: "data-button-searchid", value: id + "_MapSearchButton" },
         ],
         styles: [
            { key: "width", value: `calc(100% - ${options.distance.widthPx}px)` }
         ],
         handlers: [
            {
               eventName: "change", action: function (event) {
                  let value = $(this).val();
               }
            }
         ]
      }
   );
   $(mappingSearchInputDiv).attr("id", id + "_container");
   $(mappingSearchInputDiv).attr("data-search-id", id);
   $(mappingSearchInputDiv).addClass("office-map-container")

   // Create the Distance Selector
   let distanceSelector = document.createElement("select");
   $(distanceSelector).attr("id", id + "_distance");
   $(distanceSelector).attr("aria-label", getTranslatedText("Distance", "Distance"));
   $(distanceSelector).attr("data-zoomlevel", 11);
   $(distanceSelector).addClass("form-control");
   $(distanceSelector).css("width", `${options.distance.widthPx}px`);
   $(distanceSelector).change(function () {
      let selectedValue = $(this).val();
      let option = $(this).find(`option[value="${selectedValue}"]`);
      $(this).attr("data-zoomlevel", $(option).attr("data-zoomlevel"));
   });

   if (options.distance.optionMode == "calculate") {
      for (let i = options.distance.calculate.minKm; i <= options.distance.calculate.maxKm; i += options.distance.calculate.stepKm) {
         $(distanceSelector).append(`<option value="${i}" data-zoomlevel="${options.distance.calculate.zoomLevel}">${i}km</option>`);
      }
   }
   else if (options.distance.optionMode == "options") {
      for (let i = 0; i < options.distance.options.length; i++) {
         let option = options.distance.options[i];
         $(distanceSelector).append(`<option value="${option.distance}" data-zoomlevel="${option.zoomLevel}">${option.distance}km</option>`);
      }
   }
   $(distanceSelector).val(options.distance.defaultValue);
   $(distanceSelector).change();
   $(distanceSelector).insertAfter($(mappingSearchInputDiv).find("input"))

   // Add Enter Keypress for Search
   let mappingSearchInput = $(mappingSearchInputDiv).find("input")
   mappingSearchInput.keypress(function (event) {
      if (event.which == 13) /* Enter Pressed */ {
         console.log(event);

         // Only trigger if we are not readonly
         if (!$(this).is("[readonly]")) {
            console.log("Enter pressed. Triggering search");
            let searchBtnId = $(this).attr("data-button-searchid");
            $("#" + searchBtnId).click();
         }
      }
   });

   // Create the Position Button
   let mappingInputGroup = createInputGroup(id, {
      element: $(mappingSearchInputDiv).find("input"),
      buttons: [
         {
            content: {
               glyphicon: {
                  type: "glyphicon-map-marker",
                  aria: getTranslatedText("Use Location", "Utiliser l'emplacement"),
                  role: "button"
               }
            },
            buttonType: "btn-default",
            position: "left",
            attributes: [
               { key: "id", value: id + "_RequestLocationButton" },
               { key: "title", value: getTranslatedText("Use Location", "Utiliser l'emplacement") },
               { key: "data-prompted", value: false },
               { key: "data-latitude", value: "" },
               { key: "data-longitude", value: "" },
               { key: "data-accuracy", value: "" },
               { key: "data-backingId", value: id },
               { key: "data-mapid", value: id + "_map" },
               { key: "data-map-enabled", value: variableExists(options.mapSettings) }
            ],
            action: function () {
               let backingId = $(this).attr("data-backingId");
               let mapId = $(this).attr("data-mapid");
               $(this).attr("data-prompted", true);

               try {
                  let geolocation = navigator.geolocation;
                  if (variableExists(geolocation)) {
                     geolocation.getCurrentPosition((position) => {
                        sessionSaveObject("browser-location-position", position);
                        let crd = position.coords;
                        let searchQuery = `${crd.longitude},${crd.latitude}`;

                        console.log('Your current position is:');
                        console.log(`Latitude : ${crd.latitude}`);
                        console.log(`Longitude: ${crd.longitude}`);
                        console.log(`More or less ${crd.accuracy} meters.`);

                        $(this).attr("data-latitude", crd.latitude);
                        $(this).attr("data-longitude", crd.longitude);
                        $(this).attr("data-accuracy", crd.accuracy);

                        let backingField = $("#" + backingId);
                        backingField.val(searchQuery);
                        backingField.attr("data-searchquery-position", searchQuery);
                        backingField.attr("data-browser-latitude", crd.latitude);
                        backingField.attr("data-browser-longitude", crd.longitude);
                        backingField.attr("data-browser-accuracy", crd.accuracy);
                        backingField.trigger("onPositionChange", [{
                           type: "GPS",
                           position: position,
                           latitude: crd.latitude,
                           longitude: crd.longitude,
                           accuracy: crd.accuracy,
                           searchQuery: searchQuery,
                           eventOrigin: {
                              backingField: backingField,
                              mapId: mapId,
                              map: $("#" + mapId),
                              button: $(this)
                           }
                        }]);

                        backingField.removeAttr("readonly");
                     }, (error) => {
                        console.log("There was an error requesting Location");
                        console.log(error);

                        // Disable the Readonly Atribute and jump straight to the Search Bar
                        let backingField = $("#" + backingId);
                        backingField.removeAttr("readonly");
                     }, {
                        maximumAge: 0,
                        timeout: Infinity,
                        enableHighAccuracy: false
                     });
                  }
                  else {
                     console.log("Geolocation has been disabled by browser/organization settings");

                     // Disable the Readonly Atribute and jump straight to the Search Bar
                     let backingField = $("#" + backingId);
                     backingField.removeAttr("readonly");
                  }
               }
               catch (ex) {
                  console.log("There was an Exception requesting Location");
                  console.log(ex);

                  // Disable the Readonly Atribute and jump straight to the Search Bar
                  let backingField = $("#" + backingId);
                  backingField.removeAttr("readonly");
               }
            }
         },
         {
            content: {
               glyphicon: {
                  type: "glyphicon-search",
                  aria: getTranslatedText("Search", "Rechercher"),
                  role: "button"
               }
            },
            buttonType: "btn-default",
            position: "right",
            attributes: [
               { key: "id", value: id + "_MapSearchButton" },
               { key: "title", value: getTranslatedText("Search", "Rechercher") },
               { key: "data-backingId", value: id },
               { key: "data-mapid", value: id + "_map" },
               { key: "data-map-enabled", value: variableExists(options.mapSettings) }
            ],
            action: function () {
               let backingId = $(this).attr("data-backingId");
               let mapId = $(this).attr("data-mapid");
               let requestLocationButtonId = $("#" + backingId).attr("data-button-requestlocationid");

               let searchQuery = $("#" + backingId).val();
               if (isNullOrWhitespace(searchQuery)) { console.log("Please specify a search query"); return; }

               queryGeolocationPermission.then((result) => {
                  if (result.state == 'granted') {
                     $("#" + backingId).trigger("onSearch", [{
                        searchQuery: searchQuery,
                        mapId: mapId
                     }]);
                  }
                  else if (result.state == 'prompt') {
                     $("#" + requestLocationButtonId).click();
                  }
                  else /* permission was denied */ {
                     let wasPromptedPreviously = $("#" + requestLocationButtonId).attr("data-prompted");
                     if (wasPromptedPreviously == "true") {
                        $("#" + backingId).trigger("onSearch", [{
                           searchQuery: searchQuery,
                           mapId: mapId
                        }]);
                     }
                     else {
                        $("#" + requestLocationButtonId).click();
                     }
                  }
               }).catch(x => {
                  console.log("Error occured during query for Geolocation Permission. Treating as Denied");
                  console.log(x.error);

                  $("#" + backingId).trigger("onSearch", [{
                     searchQuery: searchQuery,
                     mapId: mapId
                  }]);
               });
            }
         }
      ]
   });

   // Check the Geolocation Permissions to Pre-enable if permission has already been prompted
   // If so, toggle the Fuzzy Search button
   // if (!isSafari()) {
   queryGeolocationPermission().then((result) => {
      if (result.state == 'prompt') { return; }
      console.log("Geolocation has already been granted/denied. Unlocking fields");

      $(mappingSearchInputDiv).find("input").removeAttr("readonly");
      $(mappingInputGroup).find("#" + id + "_RequestLocationButton").attr("data-prompted", true);
   }).catch(x => {
      console.log("Geolocation error has occured. Default: Denied. Unlocking fields");
      console.log(x.error);

      $(mappingSearchInputDiv).find("input").removeAttr("readonly");
      $(mappingInputGroup).find("#" + id + "_RequestLocationButton").attr("data-prompted", true);
   });
   // }

   // Create the Map
   let mapDiv = document.createElement("div");
   $(mapDiv).attr("id", id + "_map");
   mappingSearchInputDiv.appendChild(mapDiv);
   if (variableExists(options.mapSettings)) {
      $(mapDiv).css("width", "auto");
      $(mapDiv).css("height", options.mapSettings.height);
   }

   // Return the Search Field
   return mappingSearchInputDiv;
}

/**
 * Initializes Azure Maps
 * @param {AuthenticationOptions} authOptions 
 * @return {AuthenticationOptions} The Auth Options that were used to initialize the map 
 * @remarks https://docs.microsoft.com/en-us/javascript/api/azure-maps-control/atlas.authenticationoptions?view=azure-maps-typescript-latest
 */
function InitializeAzureMaps(authOptions) {
   if (!variableExists(authOptions)) {
      authOptions = {
         authType: "anonymous",
         clientId: $("#azureMapsClientId").val(),
         getToken: function (resolve, reject, map) {
            var tokenServiceUrl = $("#azureMapsAuthUrl").val();
            fetch(tokenServiceUrl)
               .then(r => r.text())
               .then(token => {
                  // console.log("Auth Token Retrieved: " + token);
                  let fixedToken = token.slice(1, -1);

                  resolve(fixedToken)
                  return fixedToken;
               });
         }
      };
   }

   atlas.setAuthenticationOptions(authOptions);
   return authOptions;
}

/**
 * Instantiates The Map
 * @param {HTMLDivElement} mapDiv The Map Div  
 * @param {{
 *    center: number[],
 *    zoom: number,
 * }} mapSettings Settings to generate the map with 
 * @param {{
 *    marker: number[],
 *    enableFullscreenMode: boolean
 * }} options Additional Generation Settings
 * @returns {{
 *    map: atlas.Map,
 *    popup: atlas.Popup,
 *    additional: {
 *       defaultMarker: atlas.HtmlMarker,
 *       mapDiv: HTMLDivElement
 *    }
 * }}
 * @remarks https://docs.microsoft.com/en-us/javascript/api/azure-maps-control/atlas.map?view=azure-maps-typescript-latest#markers
 */
function InstantiateMappingService(mapDiv, mapSettings, options) {
   if (!variableExists(mapSettings)) { mapSettings = {}; }
   if (!variableExists(mapSettings.center)) { mapSettings.center = [50.802599, -130.1974337]; /* Center on CANADA */ }
   if (!variableExists(mapSettings.zoom)) { mapSettings.zoom = 11; }
   if (!variableExists(mapSettings.minZoom)) { mapSettings.minZoom = 9; }
   if (!variableExists(mapSettings.maxZoom)) { mapSettings.maxZoom = 15; }
   if (!variableExists(mapSettings.language)) { mapSettings.language = getTranslatedText('en-ca', 'fr-ca'); }
   if (!variableExists(mapSettings.interactive)) { mapSettings.interactive = false; }
   if (!variableExists(mapSettings.dragPanInteraction)) { mapSettings.dragPanInteraction = false; }
   if (!variableExists(mapSettings.dragRotateInteraction)) { mapSettings.dragRotateInteraction = false; }
   if (!variableExists(options.enableFullscreenMode)) { options.enableFullscreenMode = true; }

   if (!variableExists(mapSettings.authOptions)) {
      mapSettings.authOptions = InitializeAzureMaps();
   }

   //Instantiate a map object
   let map = new atlas.Map(mapDiv, mapSettings);
   $(mapDiv).attr("data-map-initialized", true);

   let popup = new atlas.Popup();
   popup.attach(map);
   $(popup.closeBtn).attr("type", "button");

   // Attempt to enable the Fullscreen Mode
   try {
      if (options.enableFullscreenMode == true) {
         //Create an instance of the fullscreen control. By default it will toggle the map in and out of fullscreen mode. 
         let fsc = new atlas.control.FullscreenControl({
            style: 'auto',
         });

         //Optionally, add events to monitor when the fullscreen mode is entered and exited.
         map.events.add('fullscreenchanged', fsc, function (isFullscreen) {
            console.log("Map Fullscreen Toggled " + isFullscreen);
         });

         //Add the fullscreen control to the map.
         map.controls.add(fsc, {
            position: 'top-right'
         });
      }
   } catch (ex) {
      console.log("There was an error enabling Fullscreen Mode");
   }

   // Add a default Marker to the Map
   let marker = null;
   try {
      if (variableExists(options.marker)) {
         marker = new atlas.HtmlMarker({
            position: options.marker,
            color: "DimGray"
         });
         map.markers.add(marker);
      }
   } catch (ex) {
      console.log("There was an error generating the defaultMarker");
   }

   return {
      map: map,
      popup: popup,
      additional: {
         defaultMarker: marker,
         mapDiv: mapDiv
      }
   };
}

/**
 * Initializes the Search Service
 * @param {"anonymous"|"subscriptionKey"} authMode The Authentication Mode
 * @returns {{
 *    search: any,
 *    pipeline: any,
 *    subscriptionKeyCredential: any
 * }}
 */
function InitializeSearchService(authMode, map) {
   let credential;
   if (authMode == "subscriptionKey") {
      console.log("Initializing Search with Subscription Key");
      let subscriptionKey = atlas.getSubscriptionKey();
      if (isNullOrWhitespace(subscriptionKey)) {
         InitializeAzureMaps();
         subscriptionKey = atlas.getSubscriptionKey();
      }

      // Instantiate Search
      // Use SubscriptionKeyCredential with a subscription key
      var subscriptionKeyCredential = new atlas.service.SubscriptionKeyCredential(subscriptionKey);
      credential = subscriptionKeyCredential;
   }
   else if (authMode == "anonymous") {
      console.log("Initializing Search as Anonymous");
      credential = new atlas.service.MapControlCredential(map);
   }

   // Create the Pipeline
   var pipeline = atlas.service.MapsURL.newPipeline(credential);

   // Construct the SearchURL object
   var searchURL = new atlas.service.SearchURL(pipeline);

   return {
      search: searchURL,
      pipeline: pipeline,
      credential: credential
   };
}

/**
 * Calculates the Distance in Km using two points
 * @param {{latitude: number, longitude: number}} point1 The latitude/longitude pair for this coordinate
 * @param {{latitude: number, longitude: number}} point2 The latitude/longitude pair for this coordinate
 * @returns {number} The distance in kilometers
 */
function getDistanceFromLatLonInKm(point1, point2) {
   var R = 6371; // Radius of the earth in km
   var dLat = deg2rad(point2.latitude - point1.latitude);  // deg2rad below
   var dLon = deg2rad(point2.longitude - point1.longitude);
   var a =
      Math.sin(dLat / 2) * Math.sin(dLat / 2) +
      Math.cos(deg2rad(point1.latitude)) * Math.cos(deg2rad(point2.latitude)) *
      Math.sin(dLon / 2) * Math.sin(dLon / 2)
      ;
   var c = 2 * Math.atan2(Math.sqrt(a), Math.sqrt(1 - a));
   var d = R * c; // Distance in km
   return d;
}

/**
 * Converts Degrees to Radians
 * @param {number} deg The degrees value of the object
 * @returns {number} The degrees converted to Radians
 */
function deg2rad(deg) {
   return deg * (Math.PI / 180)
}

/**
 * 
 * @param {{latitude: number, longitude: number}} location The location to calculate distance from
 * @param {{
 *    maxDistanceKM?: number,
 *    selectedLanguage: string,
 *    selectedServiceOfferings: { id: string, name: string }[],
 *    selectedAdditionalOptions: { id: string, name: string }[]
 * }} options Additional options for finding the closest office
 * @returns {{
 *    id: string,
 *    name: string,
 *    nameEn: string,
 *    nameFr: string,
 *    office: any,
 *    officeFetch: any,
 *    distance: {
 *       line: number,
 *       routing: number?
 *    },
 *    location: {latitude: number, longitude: number}
 * }[]} An array containing the closest offices to your specified coordinates, sorted by distance
 * @remarks How to calculate Routing Matrix: https://github.com/Azure-Samples/AzureMapsCodeSamples/blob/master/AzureMapsCodeSamples/Spatial%20Analysis/Calculate%20nearest%20locations.html
 */
async function findClosestOffices(location, options) {
   if (!variableExists(options)) { options = {}; }
   if (!variableExists(options.maxDistanceKM)) { options.maxDistanceKM = 99999; }

   let esdc_officeadditionalservices = sessionLoadObject("esdc_officeadditionalservices");
   let esdc_offices = sessionLoadObject("esdc_offices"); //sessionLoadObject("esdc_offices/fetchxml");
   await retrieveOfficeFetch({
      displayProgress: false
   });

   // Remove all selectedAdditionalOptions that aren't usable in office search
   options.selectedAdditionalOptions = options.selectedAdditionalOptions.filter(x => {
      let service = esdc_officeadditionalservices.value.find(service => x.id.toLowerCase() == service.esdc_officeadditionalserviceid.toLowerCase());
      if (service == null) return false;
      return service.esdc_validateoffice == true;
   });

   // Filter the Offices!
   let cached_inputs = getCachedRecordInputs();
   let output = esdc_offices.value.map((x, index) => {
      let officeLocation = {
         latitude: x["esdc_latitude"],
         longitude: x["esdc_longitude"]
      };
      let lineDistance = getDistanceFromLatLonInKm(location, officeLocation);
      let officeFetch = esdc_officesFetch.find(f => f["esdc_officeid"] == x["esdc_officeid"]);

      return {
         office: x,
         officeFetch: officeFetch,
         name: getTranslatedText(x["esdc_nameenglish"], x["esdc_namefrench"]),
         nameEn: x["esdc_nameenglish"],
         nameFr: x["esdc_namefrench"],
         id: x["esdc_officeid"],
         location: officeLocation,
         distance: {
            line: lineDistance,
            routing: null
         }
      };
   })
      .filter(x => {
         if (variableExists(x.distance.routing)) { return x.distance.routing <= options.maxDistanceKM; }
         return x.distance.line <= options.maxDistanceKM;
      })
      .filter(x => {
         if (variableExists(options.selectedLanguage) && options.selectedLanguage != "564190002" /* Other */) {
            let language = x.officeFetch.esdc_availablelanguages.find(x => x.value == options.selectedLanguage);
            if (!variableExists(language)) { return false; }
         }

         // Check Passport
         // Office Types
         // let officeTypeSCC = { name: "SCC", value: "564190000", requiresPrintFlag: false };
         // let officeTypeSCCPS = { name: "SCC-PS", value: "564190001", requiresPrintFlag: false };
         // let officeTypeConsolidated = { name: "Consolidated", value: "564190002", requiresPrintFlag: true };
         if (cached_inputs.getValue("esdc_bookingtype", true) == "564190001") /* Passport */ {
            // Require the Print Flag on Offices for certain options
            if (cached_inputs.getValue("esdc_earliesttraveldaterange") == "564190002" /* 3-5 business days */) {
               if (x.office.esdc_canprintpassports != true) { return false; }
            }
         }

         // Service Offerings
         if (variableExists(options.selectedServiceOfferings)) {
            for (let i = 0; i < options.selectedServiceOfferings.length; i++) {
               let iOption = options.selectedServiceOfferings[i];
               let service = x.officeFetch.esdc_serviceofferings.find(x => x.esdc_serviceofferingid.toLowerCase() == iOption.id);
               if (!variableExists(service)) { return false; }
            }
         }

         // Additional Options
         if (variableExists(options.selectedAdditionalOptions)) {
            for (let i = 0; i < options.selectedAdditionalOptions.length; i++) {
               let iOption = options.selectedAdditionalOptions[i];
               let service = x.officeFetch.esdc_officeadditionalservices.find(x => x.esdc_officeadditionalserviceid.toLowerCase() == iOption.id);
               if (!variableExists(service)) { return false; }

            }
         }

         return true;
      })
      .sort((a, b) => {
         let aDistance = variableExists(a.distance.routing) ? a.distance.routing : a.distance.line;
         let bDistance = variableExists(b.distance.routing) ? b.distance.routing : b.distance.line;
         return aDistance - bDistance;
      })

   return output;
}

/**
 * Generates an array of Selectable Offices for the Booking Tool
 *  @param {string} idPrefix The prefix for the IDs
 *  @param {{
 *    id: string,
 *    name: string,
 *    nameEn: string,
 *    nameFr: string,
 *    office: any,
 *    officeFetch: any,
 *    distance: { line: number, routing: number? },
 *    location: {latitude: number, longitude: number}
 * }[]} offices The array of Office IDs which will be generated 
 * @param {{    
 *    backingId: string,
 *    columns: {
 *       mobile: number,
 *       tablet: number,
 *       desktop: number,
 *       desktopxl: number
 *    },
 *    selection: {
 *       type: "hidden"|"checkbox"|"radio",
 *       name: string,
 *       panelSelectedClass?: "panel-primary"|"panel-success"|"panel-info"|"panel-warning"|"panel-danger"|string,
 *       panelDeselectedClass?: "panel-primary"|"panel-success"|"panel-info"|"panel-warning"|"panel-danger"|string,
 *       hideInput?: boolean,
 *       position?: "hidden"|"panel-heading"|"panel-footer",
 *       onChange?: Function
 *    },
 *    errorSettings: {
 *       searchTerm: string,
 *       distance: number
 *    },
 *    listClass?: string,
 *    listItemClass?: string,
 *    scrollMaxHeight?: string,
 *    markerIDPosition?: "none"|"title"|"body"|"body-merge-distance"
 * }} options The number of columns to display in a row, up to a maximum of 12
 * @returns {HTMLDivElement} A row containing all of the generated Office Information
 */
function generateOfficeList(idPrefix, response, options) {
   if (isNullOrWhitespace(options.markerIDPosition)) { options.markerIDPosition = "body-merge-distance"; }
   if (isNullOrWhitespace(options.listClass)) { options.listClass = ""; }
   if (isNullOrWhitespace(options.listItemClass)) { options.listItemClass = ""; }
   if (isNullOrWhitespace(options.selection.position)) { options.selection.position = "panel-footer"; }
   if (isNullOrWhitespace(options.selection.panelSelectedClass)) { options.selection.panelSelectedClass = "panel-success"; }
   if (isNullOrWhitespace(options.selection.panelDeselectedClass)) { options.selection.panelDeselectedClass = "panel-default"; }
   if (!variableExists(options.columns)) { options.columns = {}; }
   if (!variableExists(options.columns.mobile)) { options.columns.mobile = 1; }
   if (!variableExists(options.columns.tablet)) { options.columns.tablet = 2; }
   if (!variableExists(options.columns.desktop)) { options.columns.desktop = 3; }
   if (!variableExists(options.columns.desktopxl)) { options.columns.desktopxl = 3; }

   let offices = response.Offices;

   // Calculate the Column Sizes
   let columnSizeMobile = 12 / options.columns.mobile;
   let columnSizeTablet = 12 / options.columns.tablet;
   let columnSizeDesktop = 12 / options.columns.desktop;
   let columnSizeDesktopXL = 12 / options.columns.desktopxl;

   // Iterate the offices and generate the list
   let outputColumns = [];
   for (let i = 0; i < offices.length; i++) {
      let officeMeta = offices[i];

      let esdc_enableonportal = officeMeta.Attributes.find(x => x.Key == "esdc_enableonportal")?.Value;
      if (esdc_enableonportal != true) { continue; }

      let officeNumber = officeMeta.Attributes.find(x => x.Key == "esdc_officenumber")?.Value;
      let officeNameEn = officeMeta.Attributes.find(x => x.Key == "esdc_nameenglish")?.Value;
      let officeNameFr = officeMeta.Attributes.find(x => x.Key == "esdc_namefrench")?.Value;
      let officeAddressEn = officeMeta.Attributes.find(x => x.Key == "esdc_addressenglish")?.Value;
      let officeAddressFr = officeMeta.Attributes.find(x => x.Key == "esdc_addressfrench")?.Value;

      let column = document.createElement("div");
      let columnId = idPrefix + "-" + officeMeta.Office.Id;
      $(column).attr("id", columnId);
      $(column).attr("data-backingid", options.backingId);
      $(column).attr("data-officeid", officeMeta.Office.Id);
      $(column).attr("data-officenumber", officeNumber);
      $(column).attr("data-distance-line", officeMeta.DistanceLinearKm);
      $(column).attr("data-distance-routing", officeMeta.DistanceRoutingKm);
      $(column).attr("data-position-latitude", officeMeta.Position.Latitude);
      $(column).attr("data-position-longitude", officeMeta.Position.Longitude);
      $(column).attr("data-columns-mobile", options.columns.mobile);
      $(column).attr("data-columns-tablet", options.columns.tablet);
      $(column).attr("data-columns-desktop", options.columns.desktop);
      $(column).attr("data-columns-desktopxl", options.columns.desktop);
      $(column).addClass("office-column");
      $(column).addClass("col-xs-" + columnSizeMobile);
      $(column).addClass("col-sm-" + columnSizeTablet);
      $(column).addClass("col-md-" + columnSizeDesktop);
      $(column).addClass("col-lg-" + columnSizeDesktopXL);

      // Fix Visual Glitch from Bootstrap 3: https://stackoverflow.com/a/22311212
      $(column).css("display", "flex");
      $(column).css("flex-direction", "column");

      outputColumns.push(column);

      let panel = document.createElement("div");
      $(panel).attr("id", columnId + "-panel");
      $(panel).addClass("panel");
      $(panel).addClass("office-panel");
      $(panel).addClass(options.selection.panelDeselectedClass);
      $(panel).attr("data-officeid", officeMeta.Office.Id);
      $(column).append(panel);

      //#region Panel Heading
      let panelHeading = document.createElement("div");
      $(panelHeading).attr("id", columnId + "-panel-heading");
      $(panelHeading).addClass("panel-heading");
      $(panelHeading).css("text-align", "center");
      $(panel).append(panelHeading);

      let titleCheckboxId = idPrefix + "-" + officeMeta.Office.Id + "-input";
      let titleCheckbox = document.createElement("input");
      $(titleCheckbox).attr("id", titleCheckboxId);
      $(titleCheckbox).attr("value", officeMeta.Office.Id);
      $(titleCheckbox).attr("name", options.selection.name);
      $(titleCheckbox).attr("type", options.selection.type);
      $(titleCheckbox).attr("aria-describedby", columnId + "-panel-body");
      $(titleCheckbox).attr("data-panel-class-selected", options.selection.panelSelectedClass);
      $(titleCheckbox).attr("data-panel-class-deselected", options.selection.panelDeselectedClass);
      $(titleCheckbox).attr("data-panelid", idPrefix);
      $(titleCheckbox).attr("data-backingid", options.backingId);
      $(titleCheckbox).attr("data-officeid", officeMeta.Office.Id);
      $(titleCheckbox).attr("data-officename", officeMeta.Office.Name);
      $(titleCheckbox).attr("data-officename-en", officeNameEn);
      $(titleCheckbox).attr("data-officename-fr", officeNameFr);
      $(titleCheckbox).change(function () {
         let isChecked = $(this).prop("checked");
         let panel = $(this).parents(".panel");
         let inputType = $(this).attr("type");
         let panelSelectedClass = $(this).attr("data-panel-class-selected");
         let panelDeselectedClass = $(this).attr("data-panel-class-deselected");
         let panelId = $(this).attr("data-panelid");
         let backingId = $(this).attr("data-backingid");
         let officeId = $(this).attr("data-officeid");
         let officeName = $(this).attr("data-officename");
         console.log(`Checkbox Selected == esdc_office(${officeId}) - ${officeName}`);

         // Set the selected value on the Panel
         $("#" + panelId).attr("data-selected-office-id", officeId);
         $("#" + panelId).attr("data-selected-office-name", officeName);
         $("#" + panelId).attr("data-selected-office-entityname", "esdc_office");

         // If we have Radio Buttons or Hidden, then deactivate all the panels
         if (inputType == "radio" || inputType == "hidden") {
            let allOfficePanels = $(".office-panel");
            allOfficePanels.addClass(panelDeselectedClass);
            allOfficePanels.removeClass(panelSelectedClass);
         }

         // Add in the proper styles for the selection
         if (isChecked == true) {
            panel.removeClass(panelDeselectedClass);
            panel.addClass(panelSelectedClass);
         }
         else {
            panel.addClass(panelDeselectedClass);
            panel.removeClass(panelSelectedClass);
         }

         // Trigger the onOfficeSelected on the Panel
         $("#" + panelId).trigger("onOfficeSelected", [{
            backingId: backingId,
            office: {
               id: officeId,
               name: officeName,
               entityName: "esdc_office"
            }
         }]);
      });

      if (variableExists(options.selection.onChange)) {
         $(titleCheckbox).change(options.selection.onChange);
      }

      let titleLabel = document.createElement("label");
      $(titleLabel).attr("id", titleCheckboxId + "label");
      $(titleLabel).attr("for", titleCheckboxId);
      $(titleLabel).addClass("panel-title");
      $(titleLabel).css("margin-left", "0.5em");
      $(titleLabel).text((options.markerIDPosition == "title" ? (i + 1) + " - " : "") + getTranslatedText(officeNameEn, officeNameFr));

      // Determine Whether to hide or put it in the header... footer will be checked later
      // If this is going into the Panel Heading or its hidden, put it into the Heading section. 
      // If its Hidden or Footer, insert a standard <p> title
      if (options.selection.position == "hidden") {
         $(titleLabel).addClass("sr-only");
         $(titleCheckbox).addClass("sr-only");
      }

      if (options.selection.position == "panel-heading" || options.selection.position == "hidden") {
         $(panelHeading).append(titleCheckbox);
         $(panelHeading).append(titleLabel);
      }

      if (options.selection.position == "hidden" || options.selection.position == "panel-footer") {
         let titleP = document.createElement("p");
         $(titleP).attr("id", titleCheckboxId + "plabel");
         $(titleP).attr("data-inputid", titleCheckboxId);
         $(titleP).addClass("panel-title");
         $(titleP).html("<strong>" + $(titleLabel).text() + "</strong>");
         $(titleP).click(function () {
            let inputId = $(this).attr("data-inputid");
            $("#" + inputId).click();
         })
         $(panelHeading).append(titleP);
      }
      //#endregion

      //#region Panel Body
      let panelBody = document.createElement("div");
      $(panelBody).attr("id", columnId + "-panel-body");
      $(panelBody).addClass("panel-body");
      $(panel).append(panelBody);

      // Marker Number
      let pMarkerNumber = document.createElement("p");
      $(pMarkerNumber).append(createGlyphicon("glyphicon-map-marker", {
         title: getTranslatedText("Map Marker", "Marqueur de carte"),
         ariaLabel: getTranslatedText("Map Marker", "Marqueur de carte"),
         css: [{ key: "padding-right", value: "0.5em" }]
      }));
      $(pMarkerNumber).append(document.createTextNode((i + 1).toString()));
      if (options.markerIDPosition == "body" || options.markerIDPosition == "body-merge-distance") { $(panelBody).append(pMarkerNumber); }

      // Distance
      let pDistance = document.createElement(options.markerIDPosition == "body-merge-distance" ? "span" : "p");
      $(pDistance).append(createGlyphicon("glyphicon-road", {
         title: getTranslatedText("Distance (Kilometers)", "Distance (Kilomètres)"),
         ariaLabel: getTranslatedText("Distance (Kilometers)", "Distance (Kilomètres)"),
         css: [{ key: "padding-right", value: "0.5em" }]
      }));

      let distance = variableExists(officeMeta.DistanceRoutingKm) ? officeMeta.DistanceRoutingKm : officeMeta.DistanceLinearKm;
      $(pDistance).append(document.createTextNode(distance.toFixed(2) + "km"));
      if (options.markerIDPosition == "body-merge-distance") {
         $(pDistance).css("margin-left", "0.5em");
         $(pMarkerNumber).append(pDistance);
      }
      else { $(panelBody).append(pDistance); }

      // Address
      let pAddress = document.createElement("p");
      $(pAddress).append(createGlyphicon("glyphicon-home", {
         title: "",
         ariaLabel: "",
         css: [{ key: "padding-right", value: "0.5em" }]
      }));
      $(pAddress).append($("<strong>" + getTranslatedText("Address", "Adresse") + "</strong>"));
      $(panelBody).append(pAddress);

      let pAddressText = document.createElement("p");
      let city = officeMeta.Attributes.find(x => x.Key == "esdc_city")?.Value;
      let postalCode = officeMeta.Attributes.find(x => x.Key == "esdc_postalcode")?.Value;
      if (postalCode.length == 6) { postalCode = postalCode.substring(0, 3) + " " + postalCode.substring(3, 6); }

      let addressTranslated = getTranslatedText(officeAddressEn, officeAddressFr);
      let provinceTranslated = getTranslatedText(
         officeMeta.Attributes.find(x => x.Key == "esdc_provinceterritorystate1.esdc_nameenglish")?.Value?.Value,
         officeMeta.Attributes.find(x => x.Key == "esdc_provinceterritorystate1.esdc_namefrench")?.Value?.Value,
      );

      let fullAddressText = `${addressTranslated}, ${city}, ${provinceTranslated}, ${postalCode}`;
      $(pAddressText).text(fullAddressText);
      $(pAddress).append(pAddressText);

      // Languages
      let languagesTitle = document.createElement("span");
      $(languagesTitle).append(createGlyphicon("glyphicon-comment", {
         title: "",
         ariaLabel: "",
         css: [{ key: "padding-right", value: "0.5em" }]
      }));
      $(languagesTitle).append($("<strong>" + getTranslatedText("Languages", "Langues") + "</strong>"));

      let languages = document.createElement("div");
      let languagesUl = document.createElement("ul");
      $(languagesUl).addClass(options.listClass);

      let esdc_availablelanguages = officeMeta.Attributes.find(x => x.Key == "esdc_availablelanguages")?.Value;
      $(languagesUl).append(esdc_availablelanguages?.map(x => {
         let li = $(`<li></li>`);
         li.attr("data-value", x?.Value);

         switch (x.Value) {
            case 564190000: li.text(getTranslatedText("English", "Anglais")); break;
            case 564190001: li.text(getTranslatedText("French", "Français")); break;
            case 564190002: li.text(getTranslatedText("Other Spoken Language", "Autre langue parlée")); break;
            case 100000000: li.text(getTranslatedText("American Sign Language (ASL)", "Langue des signes américaine (ASL)")); break;
            case 100000001: li.text(getTranslatedText("Quebec Sign Language (LSQ)", "Langue des signes québécoise (LSQ)")); break;
            default: li.text(""); break;
         }

         return li;
      }));
      $(languages).append(languagesUl);

      // Additional Services
      let additionalServicesTitle = document.createElement("span");
      $(additionalServicesTitle).append(createGlyphicon("glyphicon-list-alt", {
         title: "",
         ariaLabel: "",
         css: [{ key: "padding-right", value: "0.5em" }]
      }));
      $(additionalServicesTitle).append($("<strong>" + getTranslatedText("Additional Services", "Services supplémentaires") + "</strong>"));

      let additionalServices = document.createElement("div");
      let additionalServicesUl = document.createElement("ul");
      $(additionalServicesUl).addClass(options.listClass);

      let additionalServicesOptions = officeMeta.AdditionalServices
         .filter(x => {
            return x.Attributes.find(x => x.Key == "esdc_officeadditionalservice1.esdc_displayonofficelistportal")?.Value?.Value == true;
         }).sort((a, b) => {
            let aIndex = a.Attributes.find(x => x.Key == "esdc_officeadditionalservice1.esdc_index")?.Value?.Value;
            let bIndex = b.Attributes.find(x => x.Key == "esdc_officeadditionalservice1.esdc_index")?.Value?.Value;
            return aIndex > bIndex;
         }).map(x => {
            let tmpNameEnglish = x.Attributes.find(x => x.Key == "esdc_officeadditionalservice1.esdc_nameenglish")?.Value?.Value;
            let tmpNameFrench = x.Attributes.find(x => x.Key == "esdc_officeadditionalservice1.esdc_namefrench")?.Value?.Value;

            let li = $(`<li></li>`);
            li.attr("data-id", x.Attributes.find(x => x.Key == "esdc_office_esdc_officeadditionalservicid")?.Value);
            li.attr("data-officeid", x.Attributes.find(x => x.Key == "esdc_officeid")?.Value);
            li.attr("data-serviceid", x.Attributes.find(x => x.Key == "esdc_officeadditionalserviceid")?.Value);
            li.attr("data-nameenglish", tmpNameEnglish);
            li.attr("data-namefrench", tmpNameFrench);
            li.attr("data-index", x.Attributes.find(x => x.Key == "esdc_officeadditionalservice1.esdc_index")?.Value?.Value);
            li.text(getTranslatedText(tmpNameEnglish, tmpNameFrench));
            return li;
         });

      // Check to see if there is Telephone Interpretation Service.
      let tisServiceIndex = additionalServicesOptions.findIndex(x => {
         let serviceId = $(x).attr("data-serviceid");
         return serviceId == "1c7222df-f57a-ec11-8d20-0022486de52d";
      });

      // If its available, remove it and add it to the Languages List
      if (tisServiceIndex !== -1) {
         let tisService = additionalServicesOptions.splice(tisServiceIndex, 1);
         $(languagesUl).append(tisService);
      }

      // Add the remainder to the Additional Services
      $(additionalServicesUl).append(additionalServicesOptions);
      $(additionalServices).append(additionalServicesUl);

      // Convert Languages and additional Services to Collapsables, then return 
      languages = createCollapsable(languages, languagesTitle.innerHTML, {
         defaultState: "open",
         addBreakAfterDetails: false,
         containerType: "div"
      });
      $(panelBody).append(languages);

      additionalServices = createCollapsable(additionalServices, additionalServicesTitle.innerHTML, {
         defaultState: "closed",
         addBreakAfterDetails: false,
         containerType: "div"
      });
      $(panelBody).append(additionalServices);

      // Link
      let pLink = document.createElement("p");
      $(pLink).append(createGlyphicon("glyphicon-link", {
         title: getTranslatedText("Link to Office", "Lien au bureau"),
         ariaLabel: getTranslatedText("Link to Office", "Lien au bureau"),
         css: [{ key: "padding-right", value: "0.5em" }]
      }));
      let pLinkA = document.createElement("a");

      $(pLinkA).attr("href", `https://www.servicecanada.gc.ca/tbsc-fsco/sc-dsp.jsp?rc=${officeNumber}&lang=${getTranslatedText("eng", "fra")}`);
      $(pLinkA).attr("target", "_blank");
      $(pLinkA).text(getTranslatedText("Additional Office Details", "Détails supplémentaires du bureau"));
      $(pLink).append(pLinkA);
      $(panelBody).append(pLink);
      //#endregion

      //#region Panel Footer
      let panelFooter = document.createElement("div");
      $(panelFooter).attr("id", columnId + "-panel-footer");
      $(panelFooter).addClass("panel-footer");
      $(panel).append(panelFooter);

      if (options.selection.position == "panel-footer") {
         // Add the input to the footer and apply special CSS to make it take up the whole alrea
         $(titleCheckbox).css("display", "block");
         $(titleCheckbox).css("min-width", "100%");
         $(titleCheckbox).css("width", "100%");
         $(titleCheckbox).css("min-height", "100%");
         // $(titleCheckbox).css("height", "100%"); // Breaks Chromium for some reason
         $(panelFooter).append(titleCheckbox);

         $(titleLabel).addClass("sr-only");
         $(panelFooter).append(titleLabel);
      }
   }

   // Generate the Row containing the Columns
   let row = document.createElement("div");
   $(row).attr("id", idPrefix);
   $(row).attr("data-backingid", options.backingId);
   $(row).attr("data-selected-office-id", "");
   $(row).attr("data-id-prefix", idPrefix);
   $(row).attr("data-col-count", outputColumns.length);
   $(row).attr("data-origin-latitude", response.Origin.Latitude);
   $(row).attr("data-origin-longitude", response.Origin.Longitude);
   $(row).addClass("row");
   $(row).addClass("office-row");
   $(row).css("margin-top", "1em");
   if (!isNullOrWhitespace(options.scrollMaxHeight)) {
      $(row).css("max-height", options.scrollMaxHeight);
      $(row).css("overflow-y", "scroll");
   }
   $(row).append(outputColumns);

   // Fix a visual bug unique to Bootstrap 3 Column Wrapping in a responsive design
   // https://stackoverflow.com/a/22311212
   $(row).css("display", "flex");
   $(row).css("flex-wrap", "wrap");

   // Add in Error message for "No Offices Found"
   let noOfficesAlert = createAlert(idPrefix + "-alert", "alert-warning", getTranslatedText(
      `No Offices Found.`,
      `Aucun bureau trouvé.`
   ), outputColumns <= 0);
   $(row).append(noOfficesAlert);

   return row;
}

/**
 * Retrieves the Office Information using the FetchXML API
 * @param {{
 *    displayProgress: boolean
 * }} options Additional options for the office fetch
 * @returns 
 */
async function retrieveOfficeFetch(options) {
   // If we already have an office fetch catched, then return that
   if (arrayExistsWithLength(esdc_officesFetch, 0)) { return esdc_officesFetch; }

   // If we are already loading an Office Fetch, then wait for it to complete then return it
   if (loadingOfficeFetch) {
      console.log("esdc_offices/fetchxml already in progress - Waiting...");
      await waitUntil(() => { return loadingOfficeFetch == false; })
         .then(() => { console.log("Retrieve completed"); })
         .catch(() => { console.log("An error has occured"); });
      return esdc_officesFetch;
   }

   // If the Office Fetch isn't cached, then retrieve it
   loadingOfficeFetch = true;
   if (options.displayProgress) { $("#mappingSearch_search_loading").show(); }

   console.log("Retrieving - esdc_offices/fetchxml ...");
   esdc_officesFetch = await loadMetadata(getTranslatedText("/en", "/fr") + "/api/offices/").then((x) => {
      console.log("Retrieved - esdc_offices/fetchxml");
      loadingOfficeFetch = false;

      if (options.displayProgress) { $("#mappingSearch_search_loading").hide(); }

      return x.results;
   });

   return esdc_officesFetch;
}

/**
 * All the Initialization Logic for Step 1
 */
async function configureBookingStep1() {
   let cached_inputs = getCachedRecordInputs();

   // Override the autoscroll to field, and go to the subgrid intead
   setTimeout(() => {
      try {
         scrollToAndFocus("additionalServiceInfo", "officeadditionalservice_subgrid_create");
      } catch (ex) {
         animateScrollToElement($("#additionalServiceInfo")[0], "fast");
         $("#officeadditionalservice_subgrid_create").focus();
      }
   }, 500);

   // Create the Mapping
   let mappingSearch = createMappingStructured("mappingSearch", {
      mapSettings: {
         height: "400px"
      },
      displayLimit: {
         visible: true,
         optionMode: "options",
         defaultValue: 6,
         widthPx: 100,
         options: [
            { value: 6, text: 6 },
            { value: 9, text: 9 },
            { value: 12, text: 12 },
            { value: 15, text: 15 },
            // { value: null, text: getTranslatedText("All", "Tout") },
         ]
      },
      distance: {
         visible: false,
         optionMode: "options",
         widthPx: 100,
         options: [
            { distance: 25, zoomLevel: 11 },
            { distance: 50, zoomLevel: 11 },
            { distance: 100, zoomLevel: 10 },
            { distance: 150, zoomLevel: 10 },
            { distance: 200, zoomLevel: 9 },
            { distance: 250, zoomLevel: 9 },
         ],
         defaultValue: 25,
      }
   });
   let mappingParentTd = $("#step1_tab_section_findoffice_fieldset").find("tr").first().find("td").first();
   mappingParentTd.append(mappingSearch);

   //620636 messaging at top of office search section
   if (cached_inputs.getValue("esdc_bookingtype", true) == "564190001") /* Passport */ {
      let message1 = "";
      let message2 = "";
      if (isEnglish) {
         message1 = "Please follow this link to verify that the office you select offers the service you require to meet your travel needs: <a href='https://www.cic.gc.ca/english/passport/map/map.asp' target='_blank'>Find a passport service location near you (cic.gc.ca)</a>.";
         message2 = "<a href='https://www.canada.ca/en/immigration-refugees-citizenship/services/canadian-passports/fees/passport-canada.html' target='_blank'>Additional fees</a> may apply depending on when you need the passport.";
      } else /* French */ {
         message1 = "Veuillez suivre ce lien pour vérifier que le bureau que vous sélectionnez offre le service dont vous avez besoin en prévision de votre voyage : <a href='https://www.cic.gc.ca/francais/passeport/map/carte.asp' target='_blank'>Trouvez un point de service de passeport près de chez vous (cic.gc.ca)</a>.";
         message2 = "Des <a href='https://www.canada.ca/fr/immigration-refugies-citoyennete/services/passeports-canadiens/frais/passeport-canada.html' target='_blank'>frais supplémentaires</a> peuvent s'appliquer dépendamment de la date à laquelle vous avez besoin du passeport.";
      }

      $("#step1_tab_section_findoffice_fieldset_messaging").append($("<p>" + message1 + "</p>"));
      $("#step1_tab_section_findoffice_fieldset_messaging").append($("<p>" + message2 + "</p>"));
      $("#step1_tab_section_findoffice_fieldset_messaging").append(createAlert("passportSpecificBooking_tooltip", "alert-info",
         getTranslatedText(
            "The <a href='https://www.canada.ca/en/immigration-refugees-citizenship/services/canadian-passports/processing-times.html' target='_blank'>service standards</a> apply from the time your application is submitted at your appointment.",
            "<a href='https://www.canada.ca/fr/immigration-refugies-citoyennete/services/passeports-canadiens/delais-traitement.html' target='_blank'>Les normes de service</a> s'appliquent à partir du moment où votre demande est soumise lors de votre rendez-vous."
         ), true, {
         backingId: "passportspecificText",
         describes: {
            fields: ["mappingSearch_MapSearchButton"]
         }
      }));
   }

   if (cached_inputs.getValue("esdc_bookingtype", true) == "564190001") /* Passport */ {
      let passportNoAvailabilities = createAlert("passselectTime_error", "alert-danger", getTranslatedText(
         "Passport only error",
         "Passport only error French"
      ), false, {
         role: "alert",
         backingId: "esdc_officeid",
         describes: { fields: ["esdc_officeid"] }
      });
      $("#step1_tab_section_selecttime_fieldset_messaging").append(passportNoAvailabilities);

      let message3 = getTranslatedText(
         `If no appointments are available and you have a travel date, visit one of the 35 <a href='https://ircc.canada.ca/english/passport/map/map.asp' target='_blank'>specialized passport sites (Service Canada Centre - Passport Service)</a>. A proof of travel or need must be provided for urgent pick-up service and may be required for express pick-up service.`,
         `Si aucun rendez-vous n'est disponible et vous avez une date de voyage, visitez l'un des 35 <a href=' https://ircc.canada.ca/francais/passeport/map/carte.asp' target='_blank'> sites spécialisés dans les passeports (Centres Service Canada - Service de passeport)</a>. Une preuve de voyage ou de besoin est requise pour le service de retrait en personne urgent et pourrait être requise pour le service de retrait en personne express.`
      );

      let message4 = getTranslatedText(
         `If you do not have a travel date, visit your local <a href='https://www.servicecanada.gc.ca/tbsc-fsco/sc-hme.jsp?lang=eng' target='_blank'>Service Canada Centre</a>.`,
         `Si vous n'avez pas de date de voyage, visitez votre <a href='https://www.servicecanada.gc.ca/tbsc-fsco/sc-hme.jsp?lang=fra' target='_blank'>Centre Service Canada</a> local.`
      );

      $("#step1_tab_section_selecttime_fieldset_messaging").append($("<strong><p>" + message3 + "</p></strong>"));
      $("#step1_tab_section_selecttime_fieldset_messaging").append($("<strong><p>" + message4 + "</p></strong>"));
   }

   var nextTimeslotButton = createButton("selectTime_nextButton", "selectTime_date", {
      content: {
         text: getTranslatedText("Next Available Time", "Prochaine plage horaire disponible"),
         glyphicon: {
            position: "right",
            type: "glyphicon-step-forward",
            aria: getTranslatedText("Next Available Time", "Prochaine plage horaire disponible"),
            role: "button"
         }
      },
      elementType: "button",
      buttonType: "btn-default",
      position: "right",
      role: "button",
      type: "button",
      attributes: [
         { key: "data-date-id", value: "selectTime_date" },
         { key: "data-time-id", value: "selectTime_time" },
         { key: "data-date-index", value: 0 },
         { key: "data-time-index", value: -1 },
         { key: "data-time-length", value: -1 }
      ],
      action: function () {
         console.log("Finding Next Available Timeslot");
         let dateIndex = parseInt($(this).attr("data-date-index"));
         let timeIndex = parseInt($(this).attr("data-time-index")) + 1;

         // Retrieve all the Date Options
         let dateOptions = $("#selectTime_date").find("option:not([value=-1])");
         let currentDate = dateOptions.eq(dateIndex);
         let currentDateValue = currentDate.attr("value");

         // Retrieve the Time Slots, wrapping to the next date if overflowed
         let timeOptions = groupTimeslotsForDate(currentDateValue);
         let timeOptionKeys = Object.keys(timeOptions);
         let timeOptionsCount = timeOptionKeys.length;
         $(this).attr("data-time-length", timeOptionsCount);

         // Wrap the Time and re-retrieve the time options
         if (timeIndex >= timeOptionsCount) {
            dateIndex++;
            if (dateIndex >= dateOptions.length) { dateIndex = 0; }

            timeIndex = 0;
            currentDate = dateOptions.eq(dateIndex);
            currentDateValue = currentDate.attr("value");
            timeOptions = groupTimeslotsForDate(currentDateValue);
            timeOptionKeys = Object.keys(timeOptions);
            timeOptionsCount = timeOptionKeys.length;
            $(this).attr("data-time-length", timeOptionsCount);
         }

         let currentTimeValue = timeOptionKeys[timeIndex];
         console.log(`Next Available Time: ${currentDateValue} @ ${currentTimeValue} - dateIndex:${dateIndex}, timeIndex:${timeIndex}`);

         $(this).attr("data-date-index", dateIndex);
         $(this).attr("data-time-index", timeIndex);
         $("#selectTime_date").val(currentDate.attr("value"));
         $("#selectTime_date").change();
         $("#selectTime_time").val(currentTimeValue);
         $("#selectTime_time").change();
      }
   });
   $(nextTimeslotButton).appendTo($("#step1_tab_section_selecttime_fieldset_actions"));

   // Add Refresh Button to Select_Time section
   var refreshButton = createButton("selectTime_refreshButton", "closest-offices", {
      content: {
         glyphicon: {
            type: "glyphicon-refresh",
            aria: getTranslatedText("Refresh Time Slots", "Rafraîchir les créneaux horaires"),
            position: "left",
            role: "button"
         }
      },
      elementType: "button",
      buttonType: "btn-default",
      role: "button",
      type: "button",
      classes: [],
      action: function () {
         console.log("Refresh Button Clicked");
         $("#closest-offices").find(":checked").change();
      }
   });
   $(refreshButton).appendTo($("#step1_tab_section_selecttime_fieldset_actions"));

   // Begin Retrieving Office Metadata
   retrieveOfficeFetch({
      displayProgress: false
   });

   $("#mappingSearch").on("onPositionChange", async function (event, data) {
      console.log("Position Changed: ", data);
      let distanceId = $(this).attr("data-distance-id");
      let distanceElement = $("#" + distanceId);
      let distance = parseInt(distanceElement.val());
      let zoomLevel = parseInt(distanceElement.attr("data-zoomlevel"));

      let displayLimitId = $(this).attr("data-displaylimit-id");
      let displayLimitElement = $("#" + displayLimitId);
      let displayLimit = parseInt(displayLimitElement.val());

      let cached_inputs = getCachedRecordInputs();
      let esdc_bookingtype = cached_inputs.getValue("esdc_bookingtype", true);

      let selectedLanguage = [];
      if (cached_inputs.hasValue("esdc_languageofservice")) {
         let tmpLanguage = parseInt(cached_inputs.getValue("esdc_languageofservice"));
         if (tmpLanguage != NaN) { selectedLanguage.push(tmpLanguage); }
      }

      // Check Passport
      let requiresPassportPrintCapability = calculateBookingPassportPrintCapability(); // false

      // Inclusions for Required Services
      let includeRequiredServices = true; //(esdc_bookingtype == "564190000"); // BIO

      // Calculate the Service Offerings for this booking
      let selectedServiceOfferings = calculateBookingServiceOfferings();

      // Calculate the Additional Services
      // Remove all selectedAdditionalOptions that aren't usable in office search
      let esdc_officeadditionalservices = sessionLoadObject("esdc_officeadditionalservices");
      let selectedAdditionalOptions = calculateBookingAdditionalServices({ includeAssociatedServices: true, includeVirtualServices: true })
         .filter(x => {
            let service = esdc_officeadditionalservices.value.find(service => x.id.toLowerCase() == service.esdc_officeadditionalserviceid.toLowerCase());
            if (service == null) return false;
            return service.esdc_validateoffice == true;
         });

      // Generate the Office List
      console.log("Calculating Closest Offices");
      let closestOffices = await queryOffices({
         Position: {
            Latitude: data.latitude,
            Longitude: data.longitude
         },
         Top: (displayLimit == NaN) ? null : displayLimit,
         MaxDistanceKM: null, //(distance == NaN) ? null : distance,
         Query: {
            Languages: selectedLanguage,
            OfficeIds: [],
            ProvinceIds: [],
            ServiceOfferingIds: selectedServiceOfferings.map(x => x.id),
            AdditionalServiceIds: selectedAdditionalOptions.map(x => x.id),
            IncludeWorkstations: false,
            IncludeRequiredServices: includeRequiredServices,
            RequiresPassportPrintCapability: requiresPassportPrintCapability,
            TransferRequestOfficeLocation: 0
         }
      }, {});

      let officeList = generateOfficeList("closest-offices", closestOffices, {
         backingId: "esdc_officeid",
         scrollMaxHeight: "500px",
         selection: {
            type: "radio",
            name: "closest-offices",
            position: "panel-heading"
         },
         errorSettings: {
            searchTerm: $(this).val(),
            distance: distance
         }
      });

      $(officeList).on("onOfficeSelected", function (event, args) {
         let backingId = args.backingId;
         let selectedoffice = args.office;
         console.log(`onOfficeSelected ${selectedoffice.entityName}(${selectedoffice.id}) - ${selectedoffice.name}`);

         setLookupField(backingId, selectedoffice, true);
         loadSelectTimeSection([selectedoffice.id]);
      })

      setLookupField("esdc_officeid", null, true);
      $("#closest-offices").remove();
      $(this).append(officeList);

      // Disable the Progress Indicator
      let searchProgressId = $(this).attr("data-search-progressid");
      $("#" + searchProgressId).hide();

      // Initialize and Render the Map
      if (mapService == null) {
         let mapId = data.eventOrigin.mapId;
         $("#" + mapId).trigger("onInitializeAzureMaps", [{
            mapDiv: data.eventOrigin.map[0],
            longitude: data.longitude,
            latitude: data.latitude,
            zoomLevel: zoomLevel
         }]);
      } else {
         // Move the Camera
         mapService.map.setCamera({
            center: [data.longitude, data.latitude],
            zoom: zoomLevel
         });

         // Update the Markers
         let marker = new atlas.HtmlMarker({
            position: [data.longitude, data.latitude],
            color: "DimGray"
         });
         mapService.map.markers.clear();
         //mapService.map.markers.add(marker);
      }

      // Add the Office Markers
      for (let i = 0; i < closestOffices.Offices.length; i++) {
         let office = closestOffices.Offices[i];
         let officeNameEn = office.Attributes.find(x => x.Key == "esdc_nameenglish").Value;
         let officeNameFr = office.Attributes.find(x => x.Key == "esdc_namefrench").Value;

         let marker = new atlas.HtmlMarker({
            position: [office.Position.Longitude, office.Position.Latitude],
            text: i + 1
         });

         marker.properties = {
            "title": getTranslatedText(officeNameEn, officeNameFr),
            "data-office-id": office.Office.Id,
            "data-office-name": office.Office.Name,
            "data-distance": office.DistanceLinearKm,
            "data-listelementid": "closest-offices-" + office.Office.Id,
            "data-inputid": "closest-offices-" + office.Office.Id + "-input"
         }

         mapService.map.markers.add(marker);
         mapService.map.events.add('click', marker, async function (e) {
            let marker = e.target;
            let props = marker.properties;
            let popup = mapService.popup;
            let map = mapService.map;

            // Attempt to exit fullscreen mode
            try {
               await document.exitFullscreen().catch(ex => {
                  console.log("Error exiting fullscreen mode");
                  console.log(ex);
               });
            } catch (ex) { console.log("Error exiting fullscreen mode"); console.log(ex); }

            // Select the Office
            $("#" + props["data-inputid"]).click();

            // Configure the popup
            popup.setOptions({
               content: `<div style="padding:10px;">${props.title}</div>`,
               position: marker.getOptions().position,
               pixelOffset: [0, -18]
            });
            popup.open(map);
         });
      }
   });

   $("#mappingSearch").on("onSearch", function (event, data) {
      if (searchService == null) {
         let mapId = data.mapId;
         $("#" + mapId).trigger("onInitializeAzureMaps", [{
            mapDiv: $("#" + mapId)[0],
            longitude: -130.1974337,
            latitude: 50.802599,
            zoomLevel: 11
         }]);
      }

      let searchProgressId = $("#mappingSearch").attr("data-search-progressid");
      let searchQuery = data.searchQuery; // $(this).val();
      console.log("Searching: " + searchQuery);

      // Preform the search
      searchService.search.searchFuzzy(atlas.service.Aborter.timeout(10000), searchQuery, {
         limit: 1,
         countrySet: ["CA"],
         maxFuzzyLevel: 2
      }).then((response) => {
         console.log("Search completed");
         console.log(response);

         let searchCoords = response.results.first();
         if (searchCoords == null) {
            console.log("There were no results");
            $("#" + searchProgressId).hide();
            return;
         }

         let mapId = $("#mappingSearch").attr("data-mapid");
         $("#mappingSearch").trigger("onPositionChange", [{
            type: "Search",
            latitude: searchCoords.position.lat,
            longitude: searchCoords.position.lon,
            accuracy: 100,
            searchQuery: searchQuery,
            eventOrigin: {
               mapId: mapId,
               map: $("#" + mapId),
            }
         }]);
      }).catch((error) => {
         console.log("There was an error searching Azure Maps");
         console.log(error);

         $("#" + searchProgressId).hide();
      });
   });

   // Create a Progress Indicator for the Book Appointment API
   let bookappointment_apiindicator = createProgressIndicator("bookappointment_apiindicator", {
      type: "circle",
      indeterminate: true,
      enabled: true,
      container: {
         position: "relative",
         class: "row",
         text: getTranslatedText(
            "Please wait while your appointment is being booked...",
            "Veuillez patienter pendant la prise de rendez-vous..."
         )
      },
      style: {
         display: "inline",
         position: "relative",
         left: "0px"
      },
      circle: {
         width: 100,
         height: 100,
         radius: 40
      }
   });
   $(bookappointment_apiindicator).hide();
   $(bookappointment_apiindicator).insertBefore(".actions");

   let bookappointment_apiindicator_error = createAlert("bookappointment_apiindicator_error", "alert-danger", getTranslatedText(
      "Your booking was unable to be completed. Please try again later.",
      "Votre réservation n'a pas pu être finalisée. Veuillez réessayer ultérieurement."
   ), false, {
      role: "alert",
      backingId: "selecttime_time",
      describes: { fields: ["selecttime_time"] }
   });
   $(bookappointment_apiindicator_error).insertBefore(".actions");

   let bookappointment_apiindicator_validationerror = createAlert("bookappointment_apiindicator_validationerror", "alert-danger", getTranslatedText(
      "The time you have selected is no longer available. Please select a different time and try again.",
      "L'heure que vous avez choisi n'est plus disponible. Veuillez choisir une autre heure et réessayer."
   ), false, {
      role: "alert",
      backingId: "selecttime_time",
      describes: { fields: ["selecttime_time"] }
   });
   $(bookappointment_apiindicator_validationerror).insertBefore(".actions");

   // Unlock the "Create" Next Button when an input changes
   $("input,select,textarea").change(unlockNextButton);
   $("input,select,textarea").on("input", unlockNextButton);
}

/**
 * 
 * @returns {string} The number of business days for Passport before things happen
 */
function getPassportWithinNumberBusinessDays() {
   return $("#passportWithinNumberBusinessDays").attr("data-value");
}

/**
 * Unlocks the Next Button so that it can be selected
 */
function unlockNextButton() {
   console.log($(this).attr("id") + " Changed, unlocking button");
   $("#NextButton").removeAttr("disabled");
}

/**
 * Hides and locks the Next Button
 * @param {boolean} hide Whether we should hide the button 
 */
function hideAndlockNextButton() {
   $("#NextButton").attr("disabled", "disabled");
   $("#NextButton").hide();
}

/**
 * Loads in the Selected Time Sections
 * @param {string[]} officeIds The selected officeIds to use for loading the Select Time section 
 * @returns 
 */
async function loadSelectTimeSection(officeIds) {
   let selectTimeSection = $("#step1_tab_section_selecttime_fieldset");

   // If we dont have an Office Id selected, then hide the section and exit
   if (!arrayExistsWithLength(officeIds, 0)) {
      selectTimeSection.hide();
      return;
   }

   // Show the Select Time Section, and animate a scroll to it
   selectTimeSection.show();
   animateScrollToElement(selectTimeSection, 400);

   // Display the Progress Indicator to let the user know that bookings are being retrieved
   $("#selectTime").hide()
   $("#selectTime_error").hide();
   $("#selectTime_loading").show();

   // Clear out the old selectors
   let selectTime_Date = $("#selectTime_date");
   let selectTime_Time = $("#selectTime_time");
   selectTime_Date.children().remove();
   selectTime_Time.children().remove();

   // Reset the Next Timeslot Button
   $("#selectTime_nextButton").attr("data-date-index", 0);
   $("#selectTime_nextButton").attr("data-time-index", -1);
   $("#selectTime_nextButton").attr("data-time-length", -1);

   // Begin loading the availabilities
   BookingApi_MakeApiCall(
      "GetAvailableTimeslots",
      getRecordInformation().id,
      officeIds,
      {
         payload: {
            serviceOfferingIds: calculateBookingServiceOfferings().map(x => { return x.id; }),
            additionalServiceIds: calculateBookingAdditionalServices({
               includeAssociatedServices: false,
               includeVirtualServices: true
            }).map(x => { return x.id; }),
         },
         onSuccess: function (data, status, xhr) {
            $("#selectTime_loading").hide();
            console.log(data);
            // console.log(xhr);

            // Show/Hide the Availabilities Error message if the API returns no offices
            if (variableExists(data)) {
               if (data.length <= 0) { $("#selectTime_error").show(); return; }
            }
            else { $("#selectTime_error").show(); return; }

            $("#selectTime_error").hide();
            $("#selectTime").show();

            // Cache the API Results to Session Storage for future usage
            selectTime_OfficeAvailabilities = data;
            selectTime_Office = data[0];

            // Populate the Date Selection
            let dataKeys = Object.keys(selectTime_Office.AvailableWorkstations).map(x => {
               let date = moment.utc(x);
               return {
                  key: x,
                  momentDate: date,
                  unixTime: date.unix()
               };
            }).sort((a, b) => (a.unixTime > b.unixTime) ? 1 : ((b.unixTime > a.unixTime) ? -1 : 0));

            let dateOptions = createOptionList(dataKeys.map(x => {
               return {
                  text: x.momentDate.format("dddd MMMM Do, YYYY"),
                  value: x.key,
                  groupId: selectTime_Office.Office.Id,
                  attributes: [
                     { key: "data-office-id", value: selectTime_Office.Office.id },
                     { key: "data-date", value: x.key },
                     { key: "data-date-unix", value: x.unixTime }
                  ]
               }
            }), {
               text: getTranslatedText(
                  "Please select an available date",
                  "Veuillez sélectionner une date qui est disponible"
               )
            });

            selectTime_Date.append(dateOptions);

            // Retrieve the Calendar Events and Re-Render
            try {
               console.log("Rendering Calendar...")
               calendarApi_Date.render();
               calendarApi_Time.render();

               console.log("Updating Calendar Size...")
               calendarApi_Date.updateSize();
               calendarApi_Time.updateSize();

               console.log("Attempting retrieve calendar events...")
               calendarApi_Date.refetchEvents();
            }
            catch (ex) { console.log("Full Calendar is not initialized. Can not fetch events"); }

            selectTime_Date.change();

            // Calculate the first possible availability for IRCC
            try {
               console.log("Attempting to calculate esdc_firstdateofavailability...");
               let firstDataKey = dataKeys[0];
               let tmpWorkstations = selectTime_Office.AvailableWorkstations[firstDataKey.key];
               let firstTimeslots = [];
               for (let iWorkstation = 0; iWorkstation < tmpWorkstations.length; iWorkstation++) {
                  let tmpWorkstation = tmpWorkstations[iWorkstation];
                  let firstSlot = tmpWorkstation.AvailableTimeslots[0];
                  firstTimeslots.push(moment.utc(firstSlot.CalendarFrom));
               }

               let sortedSlots = firstTimeslots.sort((a, b) => (a.unix() > b.unix()) ? 1 : ((b.unix() > a.unix()) ? -1 : 0));
               let firstSlot = sortedSlots[0];
               setDateTimeField("esdc_firstdateofavailability", firstSlot.toDate(), true);
            }
            catch (ex) { console.log("There was an error calculating the esdc_firstdateofavailability"); }
         },
         onError: function (jqXhr, textStatus, errorMessage) {
            $("#selectTime_loading").hide();
            $("#selectTime_error").show();
            $("#selectTime").hide();
         }
      }
   )
}

/**
 * Calculates the Additional Service Offerings off of past inputs
 * @param {{
 *    includeAssociatedServices: boolean,
 *    includeVirtualServices: boolean
 * }} options Additional optoins for calculating the Additional Services
 * @returns {{id: string, name: string}[]} The Additional Services applicable to this Booking record
 */
function calculateBookingAdditionalServices(options) {
   let cached_inputs = getCachedRecordInputs();

   /**@type {{id:string, name:string}[]} */
   let additionalServices = [];

   // Include the Additional Services already associated to this record
   if (options.includeAssociatedServices == true) {
      if (cached_inputs.hasValue("officeadditionalservice_subgrid_selectedservices")) {
         let officeadditionalservice_subgrid_selectedservices = cached_inputs.getValue("officeadditionalservice_subgrid_selectedservices");
         if (!isNullOrWhitespace(officeadditionalservice_subgrid_selectedservices)) {
            additionalServices = officeadditionalservice_subgrid_selectedservices.split(",")
               .map(x => {
                  return {
                     id: x,
                     name: ""
                  }
               });
         }
      }
   }

   // Generate the Virtual Service Offerings
   if (options.includeVirtualServices == true) {
      if (cached_inputs.getValue("esdc_bookingtype", true) == "564190000") /* BIO */ {
         let privateWorkstation = { name: "Private workstation", id: "A50FB899-E294-EC11-B3FE-000D3A0A0CE5".toLowerCase() };
         if (cached_inputs.getValue("esdc_privateroomrequired") == "1" /* Yes */) { additionalServices.push(privateWorkstation); }
      }
   }

   return additionalServices;
}

/**
 * Calculates the Service Offerings applicable to this Booking based off of past inputs
 * @returns {{id: string, name: string}[]} The service offerings applicable to this Booking record
*/
function calculateBookingServiceOfferings() {
   let cached_inputs = getCachedRecordInputs();
   let esdc_bookingtype = cached_inputs.getValue("esdc_bookingtype", true);

   /**@type {{id:string, name:string}[]} */
   let serviceOfferings = [];

   // BIOmetrics
   if (esdc_bookingtype == "564190000") /* BIO */ {
      let bioServiceOffering = { name: "BIO - Biometrics", id: "b438d0fc-63ab-ea11-a812-000d3af4f03d".toLowerCase() }; // BIO - Biometrics
      serviceOfferings.push(bioServiceOffering);
   }

   // Passport
   if (esdc_bookingtype == "564190001") /* Passport */ {
      let passportPassport = { name: "PASS – Passport", id: "57AF4796-67AB-EA11-A812-000D3AF4F03D".toLowerCase() }; // PASS – Passport
      let passportRA = { name: "PASS - Passport RA", id: "376598A9-67AB-EA11-A812-000D3AF4F03D".toLowerCase() }; // PASS - Passport RA
      let passport10Days = { name: "PASS - 10-day pick-up", id: "28a4d8e9-5051-ed11-bba3-0022486dfee8".toLowerCase() }; // PASS - 10-day pick-up
      let passportAllSites = { name: "PASS - All Sites", id: "17128a3f-a950-ed11-bba3-002248d51f8b".toLowerCase() }; // PASS - All Sites

      let esdc_earliesttraveldaterange = cached_inputs.getValue("esdc_earliesttraveldaterange", true);
      let esdc_passportraapplicable = cached_inputs.getValue("esdc_passportraapplicable", true);

      if (esdc_earliesttraveldaterange == "564190000" /* No Travel Date */) { // If Null or Yes associate the PassportRA service offering
         if (esdc_passportraapplicable == "0" /* No */) { serviceOfferings.push(passportPassport); }
         else /* Yes or NULL */ { serviceOfferings.push(passportAllSites); }
      }
      else if (esdc_earliesttraveldaterange == "564190001" /* 2 business days or less */) { return serviceOfferings; }
      else if (esdc_earliesttraveldaterange == "564190002" /* 3-9 business days */) {
         serviceOfferings.push(passportPassport); // + Print Flag
      }
      else if (esdc_earliesttraveldaterange == "564190003" /* 10-25 business days */) {
         if (esdc_passportraapplicable == "0" /* No */) { serviceOfferings.push(passportPassport); }
         else /* Yes or NULL */ { serviceOfferings.push(passport10Days); }
      }
      else if (esdc_earliesttraveldaterange == "564190004" /* 25+ business days */) { //If Null or Yes associate the PassportRA service offering
         if (esdc_passportraapplicable == "0" /* No */) { serviceOfferings.push(passportPassport); }
         else /* Yes or NULL */ { serviceOfferings.push(passportAllSites); }
      }
   }

   return serviceOfferings;
}

/**
 * 
 * @returns {boolean} Whether this Booking requires an Office with Passport Print Capability
 */
function calculateBookingPassportPrintCapability() {
   let cached_inputs = getCachedRecordInputs();
   let esdc_bookingtype = cached_inputs.getValue("esdc_bookingtype", true);

   if (esdc_bookingtype == "564190001") /* Passport */ {
      let esdc_earliesttraveldaterange = cached_inputs.getValue("esdc_earliesttraveldaterange", true);
      let esdc_passportraapplicable = cached_inputs.getValue("esdc_passportraapplicable", true);

      if (esdc_earliesttraveldaterange == "564190002" /* 3-9 business days */) {
         return true;
      }
   }

   return false;
}

/**
 * Calculates how many slots this Booking takes up
 * @returns {number} The number of duration slots this booking will take up
 */
function calculateBookingDurationSlots() {
   let cached_inputs = getCachedRecordInputs();
   if (cached_inputs.getValue("esdc_bookingtype", true) == "564190000") /* BIO */ {
      let esdc_numberofapplicantsbio = cached_inputs.getValue("esdc_numberofapplicantsbio");
      return clamp(esdc_numberofapplicantsbio, 1, 4);
   }
   if (cached_inputs.getValue("esdc_bookingtype", true) == "564190001") /* Passport */ {
      let esdc_numberofapplicationspassport = cached_inputs.getValue("esdc_numberofapplicationspassport");
      return clamp(esdc_numberofapplicationspassport, 1, 4);
   }

   return 1;
}

/**
 * Used on the Confirmation Page. Creates an Add to Calendar Button for this Reservation
 * @param {*} office The Office to book an appointment on
 */
function createAddToCalendarButton(office) {
   loadStyle("https://cdn.jsdelivr.net/npm/add-to-calendar-button@1.8/assets/css/atcb.min.css");
   $.getScript("https://cdn.jsdelivr.net/npm/add-to-calendar-button@1.8", function (data, textStatus, jqxhr) {
      console.log("Creating 'Add to Calendar' Button");
      let confirmationNumber = $("#esdc_bookingnumber").val();
      let startDateTime = $("#esdc_startdatetimeofficelocal").val();
      let endDateTime = $("#esdc_enddatetimeofficelocal").val();

      let translatedName = getTranslatedText(office.esdc_nameenglish, office.esdc_namefrench);
      let translatedAddress = getTranslatedText(office.esdc_addressenglish, office.esdc_addressfrench);
      let translatedProvince = getTranslatedText(office["esdc_provinceterritorystate-esdc_nameenglish"], office["esdc_provinceterritorystate-esdc_namefrench"]);
      let fullAddress = `${translatedAddress}, ${office.esdc_city}, ${translatedProvince}, ${office.esdc_postalcode}, CANADA`;

      let cancelBookingLink = $("#cancelAppointmentLink");

      // https://jekuer.github.io/add-to-calendar-button/
      let addToCalendarBtn = $(document.createElement("div"))
         .addClass("atcb")
         .css("display", "none")
         .html(JSON.stringify(
            {
               "name": `eServiceCanada Reservation #${confirmationNumber} - ${translatedName}`,
               "description": `
                  ${translatedName} <br>
                  ${getTranslatedText("Appointment Confirmation Number", "Votre numéro de confirmation est le")}: ${confirmationNumber} <br>
                  ${getTranslatedText("Office Local Time", "Heure locale du bureau")}: ${moment(startDateTime).format("YYYY-MM-DD")} @ ${moment(startDateTime).format("HH:mm")} <br>
                  ${cancelBookingLink.text()}: [url]${cancelBookingLink[0].href}[/url] <br>
                  ${fullAddress}
               `,
               "timeZone": "currentBrowser",
               "startDate": moment(startDateTime).format("YYYY-MM-DD"),
               "endDate": moment(endDateTime).format("YYYY-MM-DD"),
               "startTime": moment(startDateTime).format("HH:mm"),
               "endTime": moment(endDateTime).format("HH:mm"),
               "location": fullAddress,
               "label": getTranslatedText("Add to Calendar", "Ajouter au calendrier"),
               "options": [
                  "Apple",
                  "Google",
                  "iCal",
                  "Microsoft365",
                  "MicrosoftTeams",
                  "Outlook.com",
                  "Yahoo"
               ],
               "iCalFileName": "eServiceCanada-Reservation"
            }
         )).appendTo($("#esdc_startdatetimeofficelocal").parents("td"));
      atcb_init();
      $("#atcb_button_0").attr("type", "button");
   });
}

/**
 * Preforms an API Call to gather the available timeslots or book an appointment.
 * @param {"GetAvailableTimeslots"|"BookAppointment"|"IsTimeslotAvailable"} mode The mode of the API Call
 * @param {string} bookingId The Id of the booking
 * @param {string[]} officeIds Ids of the offices to search for availabilities on
 * @param {{
 *    apiUrl: string?,
 *    onSuccess: Function,
 *    onError: Function,
 *    payload: {
 *       from: Date?,
 *       to: Date?,
 *       workstationIds: string[]?,
 *       scheduleIds: string[]?,
 *       additionalServiceIds: string[]?,
 *       serviceOfferingIds: string[]?
 *    }?
 * }} options Additional Options for the API Call
 */
async function BookingApi_MakeApiCall(mode, bookingId, officeIds, options) {
   if (isNullOrWhitespace(options.apiUrl)) { options.apiUrl = $("#bookingAPIUrl").val(); }
   if (isNullOrWhitespace(bookingId)) { bookingId = getRecordInformation().id; }
   if (officeIds.length <= 0) {
      console.log("You must specify an Office Id to call the API");
      return;
   }

   // Calculate the CrmOrigin
   let environmentOrigin = window.top.location.host.substring(0, window.top.location.host.indexOf("."));

   // Create the Api Payload
   let fullApiUrl = options.apiUrl + "api/" + mode;
   let apiPayload = {
      // "Mode": mode,
      "CrmOrigin": environmentOrigin,
      "BookingId": bookingId,
      "OfficeIds": officeIds//.join(",")
   };

   if (variableExists(options)) {
      if (variableExists(options.payload)) {
         if (variableExists(options.payload.from)) { apiPayload.From = options.payload.from.toISOString(); }

         let additionalParameters = {};
         additionalParameters.UserLocalTime = new Date().toISOString();
         if (variableExists(options.payload.to)) { additionalParameters.To = options.payload.to.toISOString(); }
         if (arrayExistsWithLength(options.payload.workstationIds, 0)) { additionalParameters.WorkstationIds = options.payload.workstationIds; }
         if (arrayExistsWithLength(options.payload.scheduleIds, 0)) { additionalParameters.ScheduleIds = options.payload.scheduleIds; }
         if (arrayExistsWithLength(options.payload.additionalServiceIds, 0)) { additionalParameters.BookingAdditionalServiceIds = options.payload.additionalServiceIds; }
         if (arrayExistsWithLength(options.payload.serviceOfferingIds, 0)) { additionalParameters.BookingServiceOfferingIds = options.payload.serviceOfferingIds; }
         apiPayload.AdditionalParameters = additionalParameters;
         // apiPayload.AdditionalParametersJson = JSON.stringify(additionalParameters);
      }
   }

   // Send the Request
   return new Promise((resolve, reject) => {
      let ajaxApiDefinition = {
         method: "POST",
         headers: { "CrmOrigin": environmentOrigin },
         data: JSON.stringify(apiPayload),
         contentType: 'application/json; charset=utf-8',
         dataType: 'json',
         url: fullApiUrl,
         success: function (data, status, xhr) {
            console.log("Success: BookingApi_MakeApiCall");
            let dataJson = data;
            data = convertJSONToObject(data);

            // console.log(data);
            // console.log(status);
            // console.log(xhr);

            if (variableExists(options.onSuccess)) {
               options.onSuccess(data, status, xhr);
            }

            resolve({
               data: data,
               dataJson: dataJson,
               status: status,
               xhr: xhr
            });
         },
         error: function (jqXhr, textStatus, errorMessage) {
            console.log(`Error: BookingApi_MakeApiCall - ${textStatus}: ${errorMessage}`);
            if (variableExists(options.onError)) {
               options.onError(jqXhr, textStatus, errorMessage);
            }

            reject({
               jqXhr: jqXhr,
               textStatus: textStatus,
               errorMessage: errorMessage
            });
         }
      };

      $.ajax(ajaxApiDefinition);
   });
}

/**
 * 
 * @param {{
 *    Position: {
 *       Latitude: number,
 *       Longitude: number
 *    },
 *    Top: number?,
 *    MaxDistanceKM: number?,
 *    Query: {
 *       Languages: number[],
 *       OfficeIds: Guid[],
 *       ProvinceIds: Guid[],
 *       ServiceOfferingIds: Guid[],
 *       AdditionalServiceIds: Guid[],
 *       IncludeWorkstations: boolean,
 *       IncludeRequiredServices: boolean,
 *       RequiresPassportPrintCapability: boolean
 *       TransferRequestOfficeLocation: 0|1|2
 *    }
 * }} requestBody The Search Request Body
 * @param {{
 *    apiUrl: string,
 *    record: {
 *       Id: Guid,
 *       LogicalName: string
 *    }
 * }} options? Additional API Options
 */
function queryOffices(requestBody, options) {
   if (!variableExists(options)) { options = {}; }
   if (isNullOrWhitespace(options.apiUrl)) { options.apiUrl = $("#eServiceAPIUrl").val(); }
   if (!variableExists(options.record)) { options.record = {}; }
   if (isNullOrWhitespace(options.record.Id)) {
      let recordInformation = getRecordInformation();
      if (!isNullOrWhitespace(recordInformation.id)) { options.record.Id = recordInformation.id; }
      else { options.record.Id = "00000000-0000-0000-0000-000000000000"; }
   }
   if (isNullOrWhitespace(options.record.LogicalName)) {
      let recordInformation = getRecordInformation();
      if (!isNullOrWhitespace(recordInformation.entityName)) { options.record.LogicalName = recordInformation.entityName; }
      else { options.record.LogicalName = "esdc_booking"; }
   }

   // Calculate the CrmOrigin
   let environmentOrigin = window.top.location.host.substring(0, window.top.location.host.indexOf("."));
   let fullApiUrl = options.apiUrl + "api/QueryOffices";

   // Append metadata to the Api Payload
   requestBody["CrmOrigin"] = environmentOrigin;
   requestBody["Record"] = options.record;

   return fetch(fullApiUrl, {
      url: fullApiUrl,
      method: "POST",
      headers: {
         "CrmOrigin": environmentOrigin,
         "Content-Type": 'application/json; charset=utf-8'
      },
      body: JSON.stringify(requestBody)
   }).then(async (response) => {
      console.log("Success: queryOffices");
      let responseJson = await response.json();
      let result = convertJSONToObject(responseJson);
      // console.log(result)

      return result;
   }).catch(ex => {
      console.log("Error: queryOffices", ex);
   });
}

/**
 * Creates a "Contact Us button and adds it to the page"
 */
function addContactUsButton() {
   let contactDiv = $(document.createElement('div'));
   contactDiv.attr("id", 'contactButton_area');
   contactDiv.insertAfter($(".actions"));
   contactDiv.hide()

   let contactBtn = createButton("ContactUsbutton", "", {
      elementType: "a",
      buttonType: "btn-default",
      content: {
         text: getTranslatedText("Contact Us", "Contactez-nous"),
      },
      ahref: getTranslatedText(
         "https://www.canada.ca/en/employment-social-development/corporate/contact/appointment-system.html",
         "https://www.canada.ca/fr/emploi-developpement-social/ministere/systeme-prise-rendez-vous-service-canada-avis.html"
      ),
      attributes: [
         { key: "target", value: "_blank" }
      ],
      classes: ["pull-right"]
   });
   contactDiv.append(contactBtn);

   return contactDiv;
}

/**
 * Creates a "Contact Us button and adds it to the page"
 */
function addPassportOnlyText() {
   let contactDiv = $(document.createElement('div'));
   contactDiv.attr("id", 'contactButton_area');
   contactDiv.insertAfter($(".actions"));
   contactDiv.hide()

   let passportAlert = createAlert("esdc_passportbooking_infosection", "alert-info", getTranslatedText(
      `If you are booking an appointment for a passport service, please follow this link to verify that the office you select offers the service you require to meet your travel needs: <a href="https://www.cic.gc.ca/english/passport/map/map.asp" target="_blank">Find a passport service location near you (cic.gc.ca).</a> <strong>Important:</strong> The <a href="https://www.canada.ca/en/immigration-refugees-citizenship/services/canadian-passports/processing-times.html" target="_blank">service standards</a> apply from the time your application is submitted at your appointment.<br/>Additional fees may also apply depending on when your passport is required. Please follow this link for more details: <a href="https://www.canada.ca/en/immigration-refugees-citizenship/services/canadian-passports/fees/passport-canada.html" target="_blank">Pay your passport fee in Canada - Canada.ca.</a>`,
      `Si vous prenez rendez-vous pour un service de passeport, veuillez suivre ce lien pour vérifier que le bureau que vous sélectionnez offre le service dont vous avez besoin en prévision de votre voyage : <a href="https://www.cic.gc.ca/francais/passeport/map/carte.asp" target="_blank">Trouvez un point de service de passeport près de chez vous (cic.gc.ca).</a> <strong>Important:</strong> Les <a href="https://www.canada.ca/fr/immigration-refugies-citoyennete/services/passeports-canadiens/delais-traitement.html" target="_blank">normes de service</a> s'appliquent à partir du moment où votre demande est soumise lors de votre rendez-vous.<br/>Des frais supplémentaires peuvent également s'appliquer selon le moment où votre passeport est requis. Veuillez suivre ce lien pour en savoir plus: <a href="https://www.canada.ca/fr/immigration-refugies-citoyennete/services/passeports-canadiens/frais/passeport-canada.html" target="_blank">Payer les frais exigés pour un passeport au Canada - Canada.ca.</a>`
   ), false, {
   });
   // $(passportAlert).insertBefore($("#"));
}

